package com.blackledge.david.html;

import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class CSSSpecificity {

	private static final String identifierRegex = "[\\w_-]+";
	private static final Pattern id = Pattern.compile("#"+identifierRegex);
	private static final Pattern attr = Pattern.compile("\\[[^\\]]*\\]");
	private static final Pattern realClass = Pattern.compile("\\."+identifierRegex);
	private static final Pattern pseudoElement = Pattern.compile("\\:\\:?(first-line|first-letter|before|after)");
	private static final Pattern pseudoElementAny = Pattern.compile("\\:\\:"+identifierRegex);
	private static final Pattern realElement = Pattern.compile("\\w+(\\|\\w+)?");
	private static final Pattern pseudoClass= Pattern.compile("\\:"+identifierRegex);
	private static final Pattern notPseudoClass = Pattern.compile(Pattern.quote(":not"));
	
	private final int[] specificity;
	
	private CSSSpecificity(int[] specificity) {
		this.specificity = specificity;
	}
	
	public CSSSpecificity(String selector) {
		// we don't include "a" value in this since it's specific to inline where there is no selector
		int b = 0;
		int c = 0;
		int d = 0;
		
		StringBuffer reducedSelector = new StringBuffer();

		// eat :not - its arguments count, but it doesn't.
		Matcher nots = notPseudoClass.matcher(selector);
		while(nots.find()) {
//			debugOut("remove "+nots.group());
			nots.appendReplacement(reducedSelector, " ");
		}
		nots.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count #ids
		Matcher ids = id.matcher(selector);
		while(ids.find()) {
//			debugOut("count id b "+ids.group());
			++b;
			ids.appendReplacement(reducedSelector, " ");
		}
		ids.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count [attributes]
		Matcher attrs = attr.matcher(selector);
		while(attrs.find()) {
//			debugOut("count attr c "+attrs.group());
			++c;
			attrs.appendReplacement(reducedSelector, " ");
		}
		attrs.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count .classes
		Matcher classes = realClass.matcher(selector);
		while(classes.find()) {
//			debugOut("count class c "+classes.group());
			++c;
			classes.appendReplacement(reducedSelector, " ");
		}
		classes.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count ::pseudo-elements (or :pseudo-elements)
		Matcher pelements = pseudoElement.matcher(selector);
		while(pelements.find()) {
//			debugOut("count pe1 d "+pelements.group());
			++d;
			pelements.appendReplacement(reducedSelector, " ");
		}
		pelements.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();
		
		// count ::other-pseudo-elements
		Matcher pelementAnys = pseudoElementAny.matcher(selector);
		while(pelementAnys.find()) {
//			debugOut("count pe2 d "+pelementAnys.group());
			++d;
			pelementAnys.appendReplacement(reducedSelector, " ");
		}
		pelementAnys.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count :pseudo-classes
		Matcher pseudoClasses = pseudoClass.matcher(selector);
		while(pseudoClasses.find()) {
//			debugOut("count pc c "+pseudoClasses.group());
			++c;
			pseudoClasses.appendReplacement(reducedSelector, " ");
		}
		pseudoClasses.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		// count ELEMENTS
		Matcher elements = realElement.matcher(selector);
		while(elements.find()) {
//			debugOut("count real d "+elements.group());
			++d;
			elements.appendReplacement(reducedSelector, " ");
		}
		elements.appendTail(reducedSelector);
		selector = reducedSelector.toString();
		reducedSelector = new StringBuffer();

		specificity = new int[] {b, c, d};
		
//		debugOut("specificity: "+toString()+"remaining selector: "+selector);

	}
		
	private static void debugOut(Object object) {
		System.out.println(object);
	}
	
	private static final Pattern specificityRef = Pattern.compile(";/\\*([^\\*]+)\\*/;");
	
	/**
	 * Add style properties to an existing set of properties based on this
	 * specificity. Encodes the specificity in comments in the properties to
	 * assist in further additions. More specific content is later in the
	 * resulting string. If oldProps is a never-before-concatenated string, it
	 * is assumed to be inline and stays the latest in the string.
	 * 
	 * @param oldProps
	 * @param specifiedProps
	 *            the properties associated with this specificity that are to be
	 *            to added
	 * @return the combined properties with comment-encoded specificity
	 *         information.
	 */
	public String concatenateProperties(String oldProps, String specifiedProps) {
		CSSSpecificity newSpecificity = this;
		
		if(specifiedProps == null || specifiedProps.trim().length() == 0) {
			return oldProps;
		}
		specifiedProps = specifiedProps.trim();
		
		// keep it clean... we'll add a ; if we need it later
		while(specifiedProps.endsWith(";")) {
			specifiedProps = specifiedProps.substring(0, specifiedProps.length()-1);
		}
		specifiedProps = ";/*"+newSpecificity+"*/;"+specifiedProps;
		
		if(oldProps == null || oldProps.trim().length() == 0) {
			return specifiedProps+";/*inline*/;";
		}
		oldProps = oldProps.trim();
		
		Matcher matcher = specificityRef.matcher(oldProps);
		StringBuffer concatenated = new StringBuffer();
		while (matcher.find()) {
			CSSSpecificity entryspec = CSSSpecificity.fromString( matcher.group(1) );
			if (specifiedProps != null) {
				if (entryspec != null
						&& entryspec.moreSpecificThan(newSpecificity)) {
					// good place to insert.
					matcher.appendReplacement(concatenated, specifiedProps + matcher.group());
					specifiedProps = null;
				} else if (matcher.group(1).equals("inline")) {
					matcher.appendReplacement(concatenated, specifiedProps + matcher.group());
					specifiedProps = null;
				} else {
					matcher.appendReplacement(concatenated, matcher.group());
				}
			} else {
				//TODO couldn't I just do matcher.appendTail;break; ?
				// newProp already placed, just fill everything back in
				matcher.appendReplacement(concatenated, matcher.group());
			}
		}
		if (specifiedProps != null) {
			// anything remaining is the inline style attribute which is always
			// more specific than the new value we're adding.
			concatenated.append(specifiedProps);
			specifiedProps = null;
			concatenated.append(";/*inline*/;");
		}
		matcher.appendTail(concatenated);
		return concatenated.toString();
	}

	public String toString() {
		return "[" + specificity[0] + "," + specificity[1] + ","
				+ specificity[2] + "]";
	}
	
	/**
	 * Reverses the encoding from {@link #toString()}
	 * @param stringSpecificity
	 */
	public static CSSSpecificity fromString(String stringSpecificity) {
		if (stringSpecificity.startsWith("[")
				&& stringSpecificity.endsWith("]")) {
			String[] specifs = stringSpecificity.substring(1,
					stringSpecificity.length() - 1).split(",");
			if (specifs.length == 3) {
				try {
					return new CSSSpecificity(new int[] {
							Integer.parseInt(specifs[0]),
							Integer.parseInt(specifs[1]),
							Integer.parseInt(specifs[2]) });
				} catch (Exception e) {
					return null;
				}
			}
		}
		return null;
	}

	@Override
	public int hashCode() {
		return specificity.hashCode();
	}
	
	@Override
	public boolean equals(Object obj) {
		if(obj instanceof CSSSpecificity) {
			return Arrays.equals(((CSSSpecificity) obj).specificity, specificity);
		} else {
			return false;
		}
	}
	
	public boolean moreSpecificThan(CSSSpecificity newSpecificity) {
		if(newSpecificity.specificity[0] < specificity[0]) {
			return true;
		} else if (newSpecificity.specificity[0] == specificity[0]) {
			if(newSpecificity.specificity[1] < specificity[1]) {
				return true;
			} else if (newSpecificity.specificity[1] == specificity[1]) {
				if(newSpecificity.specificity[2] < specificity[2]) {
					return true;
				}
			}
		}
		return false;
	}
}
