package com.blackledge.david.html;

import java.awt.Color;
import java.lang.reflect.Field;
import java.util.HashMap;

/**
 * Handles all w3c color specifications (basic and extended names, #f0f0f0,
 * #fff, rgb(0,1,2), rgba(100%,40%,10%,0.4), hsl(300,50%,24%), etc.)
 */
public class ColorHelper {

	/**
	 * Handles all w3c color specifications (basic and extended names, #f0f0f0,
	 * #fff, rgb(0,1,2), rgba(100%,40%,10%,0.4), hsl(300,50%,24%), etc.) But
	 * this method returns null if transparency is not 100% opaque.
	 * 
	 * @param colorString
	 * @return the color, or null if it can't be parsed or has any amount of
	 *         transparency
	 */
	public static Color getColor(String colorString) {
		// standard is to consider it unparseable if system can't use alpha in colors
		return getColor(colorString, false);
	}
	
	public static Color getColor(String colorString, boolean allowTransparency) {
		if (colorString == null || colorString.trim().length() == 0) {
			return null;
		}
		colorString = colorString.trim().toLowerCase();
		if(colorString.equals("transparent")) {
			// standard is to consider it unparseable if system can't use alpha in colors
			if(!allowTransparency) {
				return null;
			}
			colorString = "rgba(0,0,0,0)";
		}
		if(colorString.equals("currentcolor")) {
			//TODO supposed to return the value of the current "color" property... need to handle outside this method.
			return null;
		}
		try {
			if (colorString.startsWith("#")) {
				return getHashColor(colorString);
			} else if (colorString.startsWith("rgb")
					|| colorString.startsWith("hsl")) {
				return getFunctionColor(colorString, allowTransparency);
			} else {
				Color namedColor = basicColors.get(colorString);
				if (namedColor != null)
					return namedColor;
				
				namedColor = extendedColors.get(colorString);
				if (namedColor != null)
					return namedColor;

				//don't even bother with Color Fields... basic/extended are all that matters
//				namedColor = getFieldColor(colorString);
//				if (namedColor != null)
//					return namedColor;
			}
			// no, don't want to cause trouble by accidentally getting a color when we expect a null.
//			// last chance... maybe it's just a dumb hex number without a leading hash
//			return getHashColor("#" + colorString);
			return null;
		} catch (Exception e) {
			return null;
		}
	}

	private static Color getFieldColor(String colorString) {
		try {
			// TODO be more robust... e.g. threw on "LIGHTGREEN"
			// probably
			// "lightGreen" could have been changed to "LIGHT_GREEN"
			// even
			// though it might not have worked anyhow.
			Field field = Color.class.getField(colorString
					.toUpperCase());
			Color color = (Color) field.get(null);
			return color;
		} catch (NoSuchFieldException e) {
			// sort of expected
			System.err.println(colorString
					+ ".toUpperCase() isn't a Color field");
			// System.err.println(e);
		} catch (SecurityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassCastException e) {
			// System.err.println(e);
		}
		return null;
	}

	private static Color getFunctionColor(String colorString, boolean allowTransparency) {
		int lp = colorString.indexOf("(");
		int rp = colorString.indexOf(")");
		if (lp > 0 && rp > 0) {
			String components[] = colorString.substring(lp + 1, rp)
					.trim().split(",");
			if (components.length == 3 || components.length == 4) {
				int alpha = 255;
				if (components.length == 4) {
					// don't bother looking for hsla or rgba, just assume 4 components means that's what we're doing.
					alpha = getFunctionAlpha(components[3]);
				}
				if(!allowTransparency && alpha < 255) {
					// standard is to consider it unparseable if system can't use alpha in colors
					return null;
				}
				if (colorString.startsWith("rgb")) {
					int c1 = getFunctionComponent(components[0], 255);
					int c2 = getFunctionComponent(components[1], 255);
					int c3 = getFunctionComponent(components[2], 255);
					return new Color(c1, c2, c3, alpha);
				} else {
					return hslColor(colorString, components, alpha);
				}
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	private static Color hslColor(String colorString, String components[], int alpha) {
		// hsl is angle, pecent, percent
		// we hack it to just truncate the angle at 360 (instead of
		// wrapping) and let function handle percent and convert it
		// to scale of 1000
		//					(((x mod 360) + 360) mod 360)
		int c1 = getFunctionComponent(components[0], 360);
		int c2 = getFunctionComponent(components[1], 1000);
		int c3 = getFunctionComponent(components[2], 1000);
		// convert to 0.0-1.0f for function
		float hh = c1 / 360f;
		float ss = c2 / 1000f;
		float ll = c3 / 1000f;
		
		// convert from hsl to hsb, based on http://ariya.blogspot.com/2008/07/converting-between-hsl-and-hsv.html
		
		// prevent divide by 0... if "lightness" is 0, it's black, regardless of everything else, right?
		if(ll == 0) return Color.black;
		float h = hh;
		ss *= (ll <= 0.5) ? ll : 1 - ll;
		float b = ll + ss;
		float s = 2 * ss / (ll + ss);
		
		int rgb = Color.HSBtoRGB(h, s, b);
		
		if (alpha < 255) {
			// FIXME no idea if this algorithm is valid!!
			int rgba = rgb | (alpha << 24);
			Color color = new Color(rgba, true);
//System.out.println("trying to convert "
//+ colorString + " to " + h + "," + s
//+ "," + l + "," + alpha + " result: "
//+ color);
			return color;
		} else {
			return new Color(rgb);
		}
	}

	private static Color getHashColor(String colorString) {
		if(colorString.length() == 4) {
			// allowed to do short notation that we have to double
			char char1 = colorString.charAt(1);
			char char2 = colorString.charAt(2);
			char char3 = colorString.charAt(3);
			colorString = "#" + char1 + char1 + char2 + char2 + char3 + char3;
		}
		return Color.decode("0x" + colorString.substring(1).trim());
	}

	private static int getFunctionAlpha(String component) {
		int alpha;
		// alpha in hsla/rgba is a 0.0-1.0 range.
		double alphad = Double.parseDouble(component.trim());
		alphad = Math.min(Math.max(alphad, 0.0), 255.0);
		alpha = (int) (255.0*alphad);
		return alpha;
	}

	private static int getFunctionComponent(String string, int max) {
		string = string.trim();
		int value;
		//could be a percentage
		if (string.endsWith("%")) {
			String percentString = string.substring(0, string.length() - 1);
			int percentage = Integer.parseInt(percentString.trim());
			float perc = percentage / 100f;
			value = (int) (((float)max) * perc);
		} else {
			value = Integer.parseInt(string);
		}
		// truncate bad values.
		value = Math.min(Math.max(value, 0), max);
		return value;
	}

	/**
	 * http://www.w3.org/TR/css3-color/#colorunits
	 */
	private static final HashMap<String, Color> basicColors = new HashMap<String, Color>();
	static {
		basicColors.put("black", new Color(0, 0, 0));
		basicColors.put("silver", new Color(192, 192, 192));
		basicColors.put("gray", new Color(128, 128, 128));
		basicColors.put("white", new Color(255, 255, 255));
		basicColors.put("maroon", new Color(128, 0, 0));
		basicColors.put("red", new Color(255, 0, 0));
		basicColors.put("purple", new Color(128, 0, 128));
		basicColors.put("fuchsia", new Color(255, 0, 255));
		basicColors.put("green", new Color(0, 128, 0));
		basicColors.put("lime", new Color(0, 255, 0));
		basicColors.put("olive", new Color(128, 128, 0));
		basicColors.put("yellow", new Color(255, 255, 0));
		basicColors.put("navy", new Color(0, 0, 128));
		basicColors.put("blue", new Color(0, 0, 255));
		basicColors.put("teal", new Color(0, 128, 128));
		basicColors.put("aqua", new Color(0, 255, 255));
	}

	/**
	 * SVG / X11 colors http://www.w3.org/TR/css3-color/#svg-color
	 */
	private static final HashMap<String, Color> extendedColors = new HashMap<String, Color>();
	static {
		extendedColors.put("aliceblue", new Color(240, 248, 255));
		extendedColors.put("antiquewhite", new Color(250, 235, 215));
		extendedColors.put("aqua", new Color(0, 255, 255));
		extendedColors.put("aquamarine", new Color(127, 255, 212));
		extendedColors.put("azure", new Color(240, 255, 255));
		extendedColors.put("beige", new Color(245, 245, 220));
		extendedColors.put("bisque", new Color(255, 228, 196));
		extendedColors.put("black", new Color(0, 0, 0));
		extendedColors.put("blanchedalmond", new Color(255, 235, 205));
		extendedColors.put("blue", new Color(0, 0, 255));
		extendedColors.put("blueviolet", new Color(138, 43, 226));
		extendedColors.put("brown", new Color(165, 42, 42));
		extendedColors.put("burlywood", new Color(222, 184, 135));
		extendedColors.put("cadetblue", new Color(95, 158, 160));
		extendedColors.put("chartreuse", new Color(127, 255, 0));
		extendedColors.put("chocolate", new Color(210, 105, 30));
		extendedColors.put("coral", new Color(255, 127, 80));
		extendedColors.put("cornflowerblue", new Color(100, 149, 237));
		extendedColors.put("cornsilk", new Color(255, 248, 220));
		extendedColors.put("crimson", new Color(220, 20, 60));
		extendedColors.put("cyan", new Color(0, 255, 255));
		extendedColors.put("darkblue", new Color(0, 0, 139));
		extendedColors.put("darkcyan", new Color(0, 139, 139));
		extendedColors.put("darkgoldenrod", new Color(184, 134, 11));
		extendedColors.put("darkgray", new Color(169, 169, 169));
		extendedColors.put("darkgreen", new Color(0, 100, 0));
		extendedColors.put("darkgrey", new Color(169, 169, 169));
		extendedColors.put("darkkhaki", new Color(189, 183, 107));
		extendedColors.put("darkmagenta", new Color(139, 0, 139));
		extendedColors.put("darkolivegreen", new Color(85, 107, 47));
		extendedColors.put("darkorange", new Color(255, 140, 0));
		extendedColors.put("darkorchid", new Color(153, 50, 204));
		extendedColors.put("darkred", new Color(139, 0, 0));
		extendedColors.put("darksalmon", new Color(233, 150, 122));
		extendedColors.put("darkseagreen", new Color(143, 188, 143));
		extendedColors.put("darkslateblue", new Color(72, 61, 139));
		extendedColors.put("darkslategray", new Color(47, 79, 79));
		extendedColors.put("darkslategrey", new Color(47, 79, 79));
		extendedColors.put("darkturquoise", new Color(0, 206, 209));
		extendedColors.put("darkviolet", new Color(148, 0, 211));
		extendedColors.put("deeppink", new Color(255, 20, 147));
		extendedColors.put("deepskyblue", new Color(0, 191, 255));
		extendedColors.put("dimgray", new Color(105, 105, 105));
		extendedColors.put("dimgrey", new Color(105, 105, 105));
		extendedColors.put("dodgerblue", new Color(30, 144, 255));
		extendedColors.put("firebrick", new Color(178, 34, 34));
		extendedColors.put("floralwhite", new Color(255, 250, 240));
		extendedColors.put("forestgreen", new Color(34, 139, 34));
		extendedColors.put("fuchsia", new Color(255, 0, 255));
		extendedColors.put("gainsboro", new Color(220, 220, 220));
		extendedColors.put("ghostwhite", new Color(248, 248, 255));
		extendedColors.put("gold", new Color(255, 215, 0));
		extendedColors.put("goldenrod", new Color(218, 165, 32));
		extendedColors.put("gray", new Color(128, 128, 128));
		extendedColors.put("green", new Color(0, 128, 0));
		extendedColors.put("greenyellow", new Color(173, 255, 47));
		extendedColors.put("grey", new Color(128, 128, 128));
		extendedColors.put("honeydew", new Color(240, 255, 240));
		extendedColors.put("hotpink", new Color(255, 105, 180));
		extendedColors.put("indianred", new Color(205, 92, 92));
		extendedColors.put("indigo", new Color(75, 0, 130));
		extendedColors.put("ivory", new Color(255, 255, 240));
		extendedColors.put("khaki", new Color(240, 230, 140));
		extendedColors.put("lavender", new Color(230, 230, 250));
		extendedColors.put("lavenderblush", new Color(255, 240, 245));
		extendedColors.put("lawngreen", new Color(124, 252, 0));
		extendedColors.put("lemonchiffon", new Color(255, 250, 205));
		extendedColors.put("lightblue", new Color(173, 216, 230));
		extendedColors.put("lightcoral", new Color(240, 128, 128));
		extendedColors.put("lightcyan", new Color(224, 255, 255));
		extendedColors.put("lightgoldenrodyellow", new Color(250, 250, 210));
		extendedColors.put("lightgray", new Color(211, 211, 211));
		extendedColors.put("lightgreen", new Color(144, 238, 144));
		extendedColors.put("lightgrey", new Color(211, 211, 211));
		extendedColors.put("lightpink", new Color(255, 182, 193));
		extendedColors.put("lightsalmon", new Color(255, 160, 122));
		extendedColors.put("lightseagreen", new Color(32, 178, 170));
		extendedColors.put("lightskyblue", new Color(135, 206, 250));
		extendedColors.put("lightslategray", new Color(119, 136, 153));
		extendedColors.put("lightslategrey", new Color(119, 136, 153));
		extendedColors.put("lightsteelblue", new Color(176, 196, 222));
		extendedColors.put("lightyellow", new Color(255, 255, 224));
		extendedColors.put("lime", new Color(0, 255, 0));
		extendedColors.put("limegreen", new Color(50, 205, 50));
		extendedColors.put("linen", new Color(250, 240, 230));
		extendedColors.put("magenta", new Color(255, 0, 255));
		extendedColors.put("maroon", new Color(128, 0, 0));
		extendedColors.put("mediumaquamarine", new Color(102, 205, 170));
		extendedColors.put("mediumblue", new Color(0, 0, 205));
		extendedColors.put("mediumorchid", new Color(186, 85, 211));
		extendedColors.put("mediumpurple", new Color(147, 112, 219));
		extendedColors.put("mediumseagreen", new Color(60, 179, 113));
		extendedColors.put("mediumslateblue", new Color(123, 104, 238));
		extendedColors.put("mediumspringgreen", new Color(0, 250, 154));
		extendedColors.put("mediumturquoise", new Color(72, 209, 204));
		extendedColors.put("mediumvioletred", new Color(199, 21, 133));
		extendedColors.put("midnightblue", new Color(25, 25, 112));
		extendedColors.put("mintcream", new Color(245, 255, 250));
		extendedColors.put("mistyrose", new Color(255, 228, 225));
		extendedColors.put("moccasin", new Color(255, 228, 181));
		extendedColors.put("navajowhite", new Color(255, 222, 173));
		extendedColors.put("navy", new Color(0, 0, 128));
		extendedColors.put("oldlace", new Color(253, 245, 230));
		extendedColors.put("olive", new Color(128, 128, 0));
		extendedColors.put("olivedrab", new Color(107, 142, 35));
		extendedColors.put("orange", new Color(255, 165, 0));
		extendedColors.put("orangered", new Color(255, 69, 0));
		extendedColors.put("orchid", new Color(218, 112, 214));
		extendedColors.put("palegoldenrod", new Color(238, 232, 170));
		extendedColors.put("palegreen", new Color(152, 251, 152));
		extendedColors.put("paleturquoise", new Color(175, 238, 238));
		extendedColors.put("palevioletred", new Color(219, 112, 147));
		extendedColors.put("papayawhip", new Color(255, 239, 213));
		extendedColors.put("peachpuff", new Color(255, 218, 185));
		extendedColors.put("peru", new Color(205, 133, 63));
		extendedColors.put("pink", new Color(255, 192, 203));
		extendedColors.put("plum", new Color(221, 160, 221));
		extendedColors.put("powderblue", new Color(176, 224, 230));
		extendedColors.put("purple", new Color(128, 0, 128));
		extendedColors.put("red", new Color(255, 0, 0));
		extendedColors.put("rosybrown", new Color(188, 143, 143));
		extendedColors.put("royalblue", new Color(65, 105, 225));
		extendedColors.put("saddlebrown", new Color(139, 69, 19));
		extendedColors.put("salmon", new Color(250, 128, 114));
		extendedColors.put("sandybrown", new Color(244, 164, 96));
		extendedColors.put("seagreen", new Color(46, 139, 87));
		extendedColors.put("seashell", new Color(255, 245, 238));
		extendedColors.put("sienna", new Color(160, 82, 45));
		extendedColors.put("silver", new Color(192, 192, 192));
		extendedColors.put("skyblue", new Color(135, 206, 235));
		extendedColors.put("slateblue", new Color(106, 90, 205));
		extendedColors.put("slategray", new Color(112, 128, 144));
		extendedColors.put("slategrey", new Color(112, 128, 144));
		extendedColors.put("snow", new Color(255, 250, 250));
		extendedColors.put("springgreen", new Color(0, 255, 127));
		extendedColors.put("steelblue", new Color(70, 130, 180));
		extendedColors.put("tan", new Color(210, 180, 140));
		extendedColors.put("teal", new Color(0, 128, 128));
		extendedColors.put("thistle", new Color(216, 191, 216));
		extendedColors.put("tomato", new Color(255, 99, 71));
		extendedColors.put("turquoise", new Color(64, 224, 208));
		extendedColors.put("violet", new Color(238, 130, 238));
		extendedColors.put("wheat", new Color(245, 222, 179));
		extendedColors.put("white", new Color(255, 255, 255));
		extendedColors.put("whitesmoke", new Color(245, 245, 245));
		extendedColors.put("yellow", new Color(255, 255, 0));
		extendedColors.put("yellowgreen", new Color(154, 205, 50));
	}
}
