package com.blackledge.david.tivo.widgets.typo;

import java.awt.Color;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import com.blackledge.david.tivo.TiVo;
import com.blackledge.david.tivo.Utility;
import com.blackledge.david.tivo.widgets.GenericList;
import com.blackledge.david.tivo.widgets.typo.TyPo.Keyboard;

import com.tivo.hme.bananas.BApplication;
import com.tivo.hme.bananas.BButton;
import com.tivo.hme.bananas.BEvent;
import com.tivo.hme.bananas.BList;
import com.tivo.hme.bananas.BScreen;
import com.tivo.hme.bananas.BView;

import com.tivo.hme.interfaces.IContext;
import com.tivo.hme.sdk.HmeEvent;
import com.tivo.hme.sdk.Resource;
import com.tivo.hme.sdk.View;
import com.tivo.hme.sdk.HmeEvent.FontInfo;
import com.tivo.hme.sdk.HmeEvent.FontInfo.GlyphInfo;

/**
 * TyPo for TiVo is an auto-completion String entry widget with an editing
 * cursor. Treat it as a BButton for setting highlights, etc. (recommended
 * bar_left/bar_right settings: -10, getWidth()+10). When the user hits SELECT
 * or ENTER, the widget will expand and reposition to open a keyboard below, and
 * auto-completions above for the current word.<BR>
 * SELECT or ENTER finalizes the entry, restores the location of the BButton,
 * hides the keyboard and completion list, and adds any previously unknown words
 * to the User Dictionary. <BR>
 * PAUSE cancels the entry, restoring the previous value and state.<BR>
 * Depending on settings, taking the cursor past the right or left edge
 * (optionally including from the keyboard area) can either finish or cancel the
 * entry.<BR>
 * The User can click UP to access the completions list, or DOWM to access the
 * keyboard. Most completions can be quickly accessed from either of those
 * states by pressing the corresponding number key.<BR>
 * The user can click LEFT or RIGHT to move a cursor within the text for editing
 * purposes. FF inserts a space, RW backspaces, while editing with the cursor
 * numbers may be entered directly, CLEAR clears the whole string (CLEAR again
 * undoes that), ADVANCE skips to the end or beginning of the string, REPLAY
 * jumps to the end of the previous word.<BR>
 * Thumbs Up or Down changes the case of the next letter while editing. When in
 * the completion or keyboard area, Thumbs Up turns on shift (Thumbs Down turns
 * off) for the next keyboard letter or capitalizes the selected completion.
 * Thumbs Up again turns on CAPS lock in the keyboard (until a thumb button is
 * pressed to turn it off or the keyboard is exited) or causes the selected
 * completion to be in all capital letters. <BR>
 * The keyboard or completion list can be exited by pressing CLEAR, CH UP/DOWN
 * (as appropriate), PAUSE, or ENTER. FF, RW, ADVANCE, and REPLAY work the same
 * as during editing.<BR>
 * The Keyboard is laid out with a first row of common letter combinations, a
 * second row of common punctuation, and the remaining rows are the alphabet
 * arranged to have vowels occupying the third row, with additional punctuation
 * and symbols to the right. Some letters are repeated in a natural position to
 * fill available space.<BR>
 * Completions use the in-jar Moby word lists "FREQ.TXT" for the 1000 most
 * frequently used words in English in order of frequency, and "COMMON.TXT" for
 * about 75k other words common to more than one dictionary. In addition, it
 * will load the following other Moby word list files if they are present in the
 * classpath: "/CROSSWD.TXT", "/CRSWD-D.TXT", "/SINGLE.TXT", "/COMPOUND.TXT",
 * "/ACRONYMS.TXT", "/NAMES.TXT", "/NAMES-M.TXT", "/NAMES-F.TXT", "/PLACES.TXT",
 * "/OFTENMIS.TXT" (in that order after COMMON.TXT). They can be found at
 * Project Gutenberg (http://gutenberg.net/) and could probably be included in
 * the classpath by simply renaming the zip to a jar and including it in the
 * classpath (the more loaded, the more memory will be used). You could override
 * the loading of FREQ.TXT and COMMON.TXT by having empty files with those names
 * earlier in your classpath than the main archive.<BR>
 * The files are loaded into the {@link WordList} singleton, so any content
 * programmatically added to it will be used.<BR>
 * After loading FREQ.TXT and before loading the other files, the
 * {@link WordList} is also populated with a User Dictionary called
 * "TyPo_for_TiVo_User_Words.txt" stored in the user's Home Directory
 * ("user.home" system property) or "/" if that is not defined. This dictionary
 * will be automatically filled with all word-like items found in the entered
 * string that aren't already part of the loaded dictionary. The file can be
 * manually edited if desired.<BR>
 * TODO User settable font, separate fonts for text vs. keyboard/completions.
 * TODO Handle LIMIT_TEXT_RENDER_WIDTH gracefully if they type too much.
 * @version $Revision$
 */
public class TyPo extends BButton {
    
	public enum Keyboard {
		DEFAULT, URL, EMAIL, PASSWORD
	}


	private static final String COMPLETION_SELECT = ">COMPLETION SELECT<";
    private static final String COMPLETION_SELECT_RIGHT = ">COMPLETION SELECT_RIGHT<";
    private static final String COMPLETION_ABORTED = ">COMPLETION ABORTED<";

    class ActiveStrategy implements KeyStrategy {

        //        activeView Strategy: 
//            up:CompletionView Strategy repeat OK (means coming from TypingView)
//            down:TypingView Strategy repeat OK (means coming from CompletionView
//            left: cursor left or "left" escape action[or just goInactive] (but bonk if repeating)
//             - non-repeat or end of repeat (maybe just on release) update completions
//            right: cursor right - bonk at end if repeating, if not repeating and at end add space or "right" escape action[or just goInactive] if at end with a space
//             - non-repeat or end of repeat (maybe just on release) update completions
//            0-9: add 0-9 repeat OK, get default completions
//            skip/replay: cursor right/left to begin/end of next/previous word (or that end of text if no more words or bonk if at end) repeat OK
//             - non-repeat or end of repeat (maybe just on release) update completions
//            clear: clear or undo clear - get default completions
//            select/enter: inactiveView Strategy
//            future: [slow/play: copy (twice: mode - move right - again to end or clear to cancel) & paste]
//            pause: inactiveView (add ", get default completions)
//            thumbs up: change case up of char after cursor, advance cursor, update completions
//            thumbs down: change case down of char after cursor, advance cursor update completions
        public boolean handleKeyPress(int code, long rawcode) {
			if(handleSlideRemoteKeyPress(code, rawcode)) {
				return true;
			}
            switch(code) {
            case KEY_SELECT:
            case KEY_ENTER:
                finishEntry();
                return true;
            case KEY_PAUSE:
                cancelEntry();
                return true;
            case KEY_UP:
            case KEY_CHANNELUP:
                if( _completionList.size() > 0 ) {
                    goCompletion();
                    return true;
                }
                return true; // always return true so left/right up down don't get forwarded.
            case KEY_DOWN:
            case KEY_CHANNELDOWN:
                goKeyboard();
                return true; // always return true so left/right up down don't get forwarded.
            case KEY_REVERSE:
                if( backspace() ) {
                    getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                    return true;
                }
                break;
            case KEY_FORWARD:
                if(type(" ")) {
					getBApp().play( TiVo.SOUND_SPEEDUP1 );
				} else {
					getBApp().play( TiVo.SOUND_BONK );
				}                	
                return true;
            case KEY_ADVANCE:
                if(advance()) {
                    getBApp().play( TiVo.SOUND_RIGHT );
                    return true;
                }
                break;
            case KEY_REPLAY:
                if(replay()) {
                    getBApp().play( TiVo.SOUND_LEFT );
                    return true;
                }
                break;
            case KEY_NUM0:
            case KEY_NUM1:
            case KEY_NUM2:
            case KEY_NUM3:
            case KEY_NUM4:
            case KEY_NUM5:
            case KEY_NUM6:
            case KEY_NUM7:
            case KEY_NUM8:
            case KEY_NUM9:
                if(type(String.valueOf((char)('0'+(code-KEY_NUM0))))) {
					getBApp().play( TiVo.SOUND_RIGHT );
				} else {
					getBApp().play( TiVo.SOUND_BONK );
				}                	
                return true;
    		case KEY_RECORD:
				if (type(".")) {
					getBApp().play(TiVo.SOUND_RIGHT);
				} else {
					getBApp().play(TiVo.SOUND_BONK);
				}
                return true;
    		case KEY_LEFT:
                if(cursorLeft()) {
                    getBApp().play( TiVo.SOUND_UPDOWN );
                    return true;
                } else if(_leftEdge != EXIT_BLOCKED) {
                    if(_leftEdge == EXIT_FINISHED) {
                        finishEntry();
                        return !_onFinishedForwardEdgeEvent;
                    } else { // EXIT_CANCEL
                        cancelEntry();
                        return !_onCancelForwardEdgeEvent;
                    }
                }
                return true; // always return true so left/right up down don't get forwarded.
            case KEY_RIGHT:
                if(cursorRight()) {
                    getBApp().play( TiVo.SOUND_UPDOWN );
                    return true;
                } else if(_rightEdge != EXIT_BLOCKED) {
                    if(_rightEdge == EXIT_FINISHED) {
                        finishEntry();
                        return !_onFinishedForwardEdgeEvent;
                    } else { // EXIT_CANCEL
                        cancelEntry();
                        return !_onCancelForwardEdgeEvent;
                    }
                }
                return true; // always return true so left/right up down don't get forwarded.
            case KEY_CLEAR:
                clear();
                return true;
            case KEY_THUMBSUP:
                if( _cursorTextPosition < getValue().length() ) {
                    String character = String.valueOf( getValue().charAt( _cursorTextPosition ) );
                    setPainting( false );
                    try {
                        if( cursorRight() && backspace() && type( character.toUpperCase() )) {
                            getBApp().play( TiVo.SOUND_SPEEDUP2 );
                        } else {
                            getBApp().play( TiVo.SOUND_BONK );
                        }
                    } finally {
                        setPainting( true );
                    }
                    return true;
                } else {
                	// treat as system shift change
                	shiftUp();
                	return true;
                }
            case KEY_THUMBSDOWN:
                if( _cursorTextPosition < getValue().length() ) {
                    String character = String.valueOf( getValue().charAt( _cursorTextPosition ) );
                    setPainting( false );
                    try {
                        if( cursorRight() && backspace() && type( character.toLowerCase() )) {
                            getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                        } else {
                        	getBApp().play( TiVo.SOUND_BONK );
                        }
                    } finally {
                        setPainting( true );
                    }
                    return true;
                } else {
                	// treat as system shift change
                    getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                    setUnshifted();
                    return true;
                }
            }
            return false;
        }

        private String _clearUndo;
        private Object _clearUndoText;
        private void clear() {
            if(getValue() == null || getValue().length() == 0) {
                if( _clearUndo != null ) {
                    // play a sound for undo clear
                    getBApp().play(TiVo.SOUND_THUMBSUP);
                    setValue( _clearUndo );
                    _clearUndo = null;
                }
            } else {
                if(_clearUndoText == null) {
                    _clearUndoText = createText( _shiftTextFont, Color.magenta.brighter(),
                            "['clear' again to undo]" );
                }
                // play a sound for clear
                getBApp().play(TiVo.SOUND_THUMBSDOWN);
                _shiftView.setResource( _clearUndoText, RSRC_TEXT_WRAP|RSRC_HALIGN_LEFT|RSRC_VALIGN_TOP );
                _clearUndo = getValue();
                setValue("");
            }
        }

        private boolean cursorRight() {
            if(_cursorTextPosition >= _value.length())
                return false;
            else
                ++_cursorTextPosition;
            updateAll();
            return true;
        }

        private boolean cursorLeft() {
            if(_cursorTextPosition <= 0)
                return false;
            else
                --_cursorTextPosition;
            updateAll();
            return true;
        }

        public boolean handleKeyRelease(int code, long rawcode) {
            switch(code) {
            case KEY_UNKNOWN:
                return handleKeyboard(rawcode);
            }
            return false;
        }

        public boolean handleKeyRepeat(int code, long rawcode) {
            switch(code) {
            case KEY_LEFT:
            case KEY_RIGHT:
            case KEY_FORWARD:
            case KEY_REVERSE:
            case KEY_REPLAY:
            case KEY_THUMBSDOWN:
            case KEY_THUMBSUP:
                return handleKeyPress( code, rawcode );
            case KEY_UNKNOWN:
                return handleKeyboard(rawcode);
            case KEY_BACKSPACE:
            	return handleKeyPress(code, rawcode);
            }
            if(code >= KEY_ASCII_START && code <= KEY_ASCII_END) {
            	return handleKeyPress(code, rawcode);
            }
            return false;
        }
    }
    class CompletionStrategy implements KeyStrategy {
//        CompletionView Strategy:
//            up/down: change list focus - if at top, bonk, if at bottom activeView Strategy
//            ch up/down: change list focus by a page - same as up/down
//            left/clear: ActiveView strategy
//            right/ff: insert completion plus space (after partial word) at cursor, update default completions, activeView Strategy
//             OR add next char from focused entry, update completions, focus same item (guaranteed to be there)
//            [rw: do backspace, update completions, stay in CompletionView strategy, focus same item else focus item 1 (or go activeView if was last char)]
//            select/enter: add completion at cursor, delete partial word before cursor, update completions, activeView Strategy
//             OR insert with a space (they can just do rw/left immediately to get further completions)
//            1-9: select list entry if visible, focus list entry if not visible, bonk if no entry
//            thumbs up: change to Title case for completions - again: change to ALL CAPS for completions - again: normal
//            thumbs down: change case state in reverse order (state initialized based on first and second char status of partial word)

      public boolean handleKeyPress(int code, long rawcode) {
			if(handleSlideRemoteKeyPress(code, rawcode))
				return true;
          switch(code) {
//          case KEY_SELECT:
//              return true;
          case KEY_CLEAR:
          case KEY_CHANNELDOWN:
          case KEY_ENTER:
          case KEY_PAUSE:
//          case KEY_LEFT:
              goActive();
              return true;
//            case KEY_ENTER:
//            _completionView.selectFocused();
//            return true;
          case KEY_DOWN:
              if(_completionList.getFocus() == _completionList.size()-1) {
                  goActive();
                  return true;
              }
              break;
          case KEY_REVERSE:
              if( backspace() ) {
                  getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                  return true;
              }
              break;
          case KEY_FORWARD:
				if (type(" ")) {
					getBApp().play(TiVo.SOUND_SPEEDUP1);
				} else {
					getBApp().play(TiVo.SOUND_BONK);
				}
				return true;
          case KEY_ADVANCE:
              if(advance()) {
                  getBApp().play( TiVo.SOUND_RIGHT );
                  return true;
              }
              break;
          case KEY_REPLAY:
              if(replay()) {
                  getBApp().play( TiVo.SOUND_LEFT );
                  return true;
              }
              break;
          case KEY_NUM1:
          case KEY_NUM2:
          case KEY_NUM3:
          case KEY_NUM4:
          case KEY_NUM5:
          case KEY_NUM6:
          case KEY_NUM7:
          case KEY_NUM8:
          case KEY_NUM9:
              int index = code-KEY_NUM1;
              index = _completionList.size()-index-1;
              if(index < _completionList.size()) {
                  _completionList.setFocus( index, true );
                  _completionView.selectFocused();
                  return true;
              }
              break;
          case KEY_THUMBSUP:
              shiftUp();
              return true;
          case KEY_THUMBSDOWN:
              getBApp().play( TiVo.SOUND_SLOWDOWN1 );
              setUnshifted();
              return true;
          }
          return false;
      }


      public boolean handleKeyRelease(int code, long rawcode) {
          switch(code) {
          case KEY_UNKNOWN:
              return handleKeyboard(rawcode);
          }
          return false;
      }

      public boolean handleKeyRepeat(int code, long rawcode) {
          switch(code) {
          case KEY_DOWN:
              if(_completionList.getFocus() == _completionList.size()-1) {
                  goActive();
                  return true;
              }
              break;
          case KEY_UNKNOWN:
              return handleKeyboard(rawcode);
          case KEY_BACKSPACE:
          	return handleKeyPress(code, rawcode);
          }
          if(code >= KEY_ASCII_START && code <= KEY_ASCII_END) {
          	return handleKeyPress(code, rawcode);
          }
          return false;
      }

  }
    class KeyboardStrategy implements KeyStrategy {
//        TypingView Strategy: 
//            up/down/left/right: move around lists - depends on current row what left/right do.
//           left at left-most: non-repeat:activeView Strategy. repeat: bonk
//           clear: activeView Strategy
//           right at right-most: if only one entry: add char(s) of entry, update completions; else bonk
//           --right at right-arrow-icon: something to replace current set with expanded set for that row.
//           --left at left-arrow-icon(from expansions only): something to restore normal set.
//           up/down: move to that row, closest index into row's list -- if in expansion, bonk at top
//           - if on first entry, activeView strategy
//           - if going to ghost entry, move to real entry
//           select/enter: add char(s) of entry, update completions, 
//            - go to activeView
//            OR stay on keyboard if in keyboard or go to completion 1 if in completions
//            (definitely stay on keyboard if no educated completions)
//           ff: insert space, update completions
//           rw: backspace, update completions
//           1-9: select completion entry, bonk if none.
//           thumbs up: Shift on - again: Shift Lock until activeView
//           thumbs down: Shift(/lock) off.
//           skip/repeat: end/begin of row.
//            

      public boolean handleKeyPress(int code, long rawcode) {
			if(handleSlideRemoteKeyPress(code, rawcode))
				return true;
          switch(code) {
          case KEY_SELECT:
//          case KEY_ENTER:
              _keyboardView.selectFocused();
              return true;
          case KEY_UP:
              getBApp().play(TiVo.SOUND_UPDOWN);
              if(!_keyboardView.up()) {
                  goActive();
              }
              return true;
          case KEY_DOWN:
              if(_keyboardView.down()) {
                  getBApp().play(TiVo.SOUND_UPDOWN);
                  return true;
              }
              return false;
          case KEY_RIGHT:
              if(_keyboardView.right()) {
                  getBApp().play(TiVo.SOUND_UPDOWN);
                  return true;
              } else if(_rightEdge != EXIT_BLOCKED) {
                  if(_rightEdge == EXIT_FINISHED && _finishedIncludesKeyboardEdge) {
                      finishEntry();
                      return !_onFinishedForwardEdgeEvent;
                  } else if(_rightEdge == EXIT_CANCEL && _cancelIncludesKeyboardEdge){
                      cancelEntry();
                      return !_onCancelForwardEdgeEvent;
                  }
              }
              return false;
          case KEY_LEFT:
              if(_keyboardView.left()) {
                  getBApp().play(TiVo.SOUND_UPDOWN);
                  return true;
              } else if(_leftEdge != EXIT_BLOCKED) {
                  if(_leftEdge == EXIT_FINISHED && _finishedIncludesKeyboardEdge) {
                      finishEntry();
                      return !_onFinishedForwardEdgeEvent;
                  } else if(_leftEdge == EXIT_CANCEL && _cancelIncludesKeyboardEdge){
                      cancelEntry();
                      return !_onCancelForwardEdgeEvent;
                  }
              }
              return false;
          case KEY_CLEAR:
          case KEY_CHANNELUP:
          case KEY_ENTER:
          case KEY_PAUSE:
              goActive();
                return true;
            case KEY_REVERSE:
                if( backspace() ) {
                    getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                    return true;
                }
                break;
            case KEY_FORWARD:
                if(type(" ")) {
                	getBApp().play( TiVo.SOUND_SPEEDUP1 );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
            case KEY_ADVANCE:
                if(advance()) {
                    getBApp().play( TiVo.SOUND_RIGHT );
                    return true;
                }
                break;
            case KEY_REPLAY:
                if(replay()) {
                    getBApp().play( TiVo.SOUND_LEFT );
                    return true;
                }
                break;
            case KEY_NUM1:
            case KEY_NUM2:
            case KEY_NUM3:
            case KEY_NUM4:
            case KEY_NUM5:
            case KEY_NUM6:
            case KEY_NUM7:
            case KEY_NUM8:
            case KEY_NUM9:
            	if(isUsingCompletions) {
	                int index = code-KEY_NUM1;
	                index = _completionList.size()-index-1;
	                if(index < _completionList.size()) {
	                    _completionList.setFocus( index, true );
	                    // passing false doesn't change mode, just updates everything
	                    _completionView.selectFocused(false);
	                    return true;
	                }
            	} else {
                    if(type(String.valueOf((char)('0'+(code-KEY_NUM0))))) {
                    	getBApp().play( TiVo.SOUND_RIGHT );
                    } else {
                    	getBApp().play( TiVo.SOUND_BONK );
                    }
                    return true;
            	}
                break;
            case KEY_THUMBSUP:
            	shiftUp();
                return true;
            case KEY_THUMBSDOWN:
                getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                setUnshifted();
                return true;                
            }
          return false;
      }

      public boolean handleKeyRelease(int code, long rawcode) {
          switch(code) {
          case KEY_UNKNOWN:
              return handleKeyboard(rawcode);
          }
          return false;
      }

      public boolean handleKeyRepeat(int code, long rawcode) {
          switch(code) {
          case KEY_LEFT:
          case KEY_RIGHT:
          case KEY_UP:
          case KEY_DOWN:
          case KEY_FORWARD:
          case KEY_REVERSE:
              return handleKeyPress( code, rawcode );
          case KEY_UNKNOWN:
              return handleKeyboard(rawcode);
          case KEY_BACKSPACE:
          	return handleKeyPress(code, rawcode);
          }
          if(code >= KEY_ASCII_START && code <= KEY_ASCII_END) {
          	return handleKeyPress(code, rawcode);
          }
          return false;
      }

  }
    class InactiveStrategy implements KeyStrategy {
//        inactiveView Strategy:
//            select/enter: activeView Strategy
//            right/left: right/left escape // handled by parent, If set up by caller.
//            up/down: up/down escape // handled by parent, If set up by caller.
//            [nothing else - leave other controls for parent app]

		public boolean handleKeyPress(int code, long rawcode) {
			if(handleSlideRemoteKeyPress(code, rawcode))
				return true;
            switch(code) {
            case KEY_SELECT:
            case KEY_ENTER:
                goActive();
                return true;
            case KEY_REVERSE:
                if( backspace() ) {
                    getBApp().play( TiVo.SOUND_SLOWDOWN1 );
                    return true;
                }
                break;
            case KEY_FORWARD:
                if(type(" ")) {
                	getBApp().play( TiVo.SOUND_SPEEDUP1 );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
            case KEY_ADVANCE:
                if(type("-")) {
                	getBApp().play( TiVo.SOUND_SPEEDUP1 );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
            case KEY_REPLAY:
                if(type(",")) {
                	getBApp().play( TiVo.SOUND_SPEEDUP1 );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
            case KEY_NUM0:
            case KEY_NUM1:
            case KEY_NUM2:
            case KEY_NUM3:
            case KEY_NUM4:
            case KEY_NUM5:
            case KEY_NUM6:
            case KEY_NUM7:
            case KEY_NUM8:
            case KEY_NUM9:
                if(type(String.valueOf((char)('0'+(code-KEY_NUM0))))) {
                	getBApp().play( TiVo.SOUND_RIGHT );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
    		case KEY_RECORD:
                if(type(".")) {
                	getBApp().play( TiVo.SOUND_RIGHT );
                } else {
                	getBApp().play( TiVo.SOUND_BONK );
                }
                return true;
            case KEY_CLEAR:
                clear();
                getBApp().play( TiVo.SOUND_SPEEDUP1 );
                return true;
            }
            return false;
        }
        private String _clearUndo;
        private void clear() {
            if(getValue() == null || getValue().length() == 0) {
                if( _clearUndo != null ) {
                    // play a sound for undo clear
                    getBApp().play(TiVo.SOUND_THUMBSUP);
                    setValue( _clearUndo );
                    _clearUndo = null;
                }
            } else {
                // play a sound for clear
                getBApp().play(TiVo.SOUND_THUMBSDOWN);
                _clearUndo = getValue();
                setValue("");
            }
        }

        public boolean handleKeyRelease(int code, long rawcode) {
            switch(code) {
            case KEY_UNKNOWN:
                return handleKeyboard(rawcode);
            }
            return false;
        }

        public boolean handleKeyRepeat(int code, long rawcode) {
            switch(code) {
            case KEY_UNKNOWN:
                return handleKeyboard(rawcode);
            case KEY_BACKSPACE:
            	return handleKeyPress(code, rawcode);
            }
            if(code >= KEY_ASCII_START && code <= KEY_ASCII_END) {
            	return handleKeyPress(code, rawcode);
            }
            return false;
        }

    }

    public interface KeyStrategy {

        boolean handleKeyRepeat(int code, long rawcode);

        boolean handleKeyRelease(int code, long rawcode);

        boolean handleKeyPress(int code, long rawcode);

    }

    CompletionView _completionView;
    KeyboardView _keyboardView;
    GenericList _completionList;
    private boolean _shifted;
    private boolean _capslock;
    private int _originalWidth = -1;
    private int _originalY = -1;
    private int _originalX = -1;
    public static final int EXIT_BLOCKED = 0;
    public static final int EXIT_CANCEL = 1;
    public static final int EXIT_FINISHED = 2;
    int _leftEdge = EXIT_BLOCKED;
    int _rightEdge = EXIT_FINISHED;
    boolean _finishedIncludesKeyboardEdge = true;
    boolean _cancelIncludesKeyboardEdge = false;
    boolean _onFinishedForwardEdgeEvent = false;
    boolean _onCancelForwardEdgeEvent = false;
    private Object _finishedAction = null;
    private Object _cancelAction = null;
    
    private class CompletionView extends BView {
        public CompletionView(BView parent, int x, int y, int width, int height, boolean visible) {
            super( parent, x, y, width, height, visible );
        }

        @Override
        public boolean handleKeyPress(int code, long rawcode) {
            if(getCurrentStrategy() != null && getCurrentStrategy().handleKeyPress( code, rawcode ))
                return true;
            if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy)
                return super.handleKeyPress( code, rawcode );
            else return false;            
        }
        @Override
        public boolean handleKeyRelease(int code, long rawcode) {
            if(getCurrentStrategy() != null && getCurrentStrategy().handleKeyRelease( code, rawcode ))
                return true;
            if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy) {
                return super.handleKeyRelease( code, rawcode );
            } else {
                return false;
            }
        }
        @Override
        public boolean handleKeyRepeat(int code, long rawcode) {
            if(getCurrentStrategy() != null && getCurrentStrategy().handleKeyRepeat( code, rawcode ))
                return true;
            if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy) {
                return super.handleKeyRepeat( code, rawcode );
            } else {
                return false;
            }
        }
        
        @Override
        public boolean handleAction(BView view, Object action) {
            if(action.equals(COMPLETION_ABORTED)) {
                goActive();
                return true;
            } else if(action.equals(COMPLETION_SELECT)
                    || action.equals(COMPLETION_SELECT_RIGHT)
            ) {
                selectFocused();
                return true;
            }
            return super.handleAction( view, action );
        }

        void selectFocused() {
            selectFocused(true);
        }
        void selectFocused(boolean thenActive) {
            // take word AFTER last space.
            BList list = _completionList;
//System.out.println("selected "+list.getFocus());
            String currentWord = getCurrentWord();
            if(list.getFocus() < 0) return;
            String selected = (String)list.get( list.getFocus() );
            if( selected == null )
                return;
            selected = selected.substring( selected.lastIndexOf( " " ) + 1 );
            if( selected.length() == 0 )
                return;
            // if this suggestion came by treating the word as lowercase, retain case.
            if( !selected.startsWith( currentWord ) ) {
                if( selected.startsWith( currentWord.toLowerCase() ) ) {
                    if( currentWord.length() > 1 && currentWord.equals( currentWord.toUpperCase() ) ) {
                        // if so far is all caps, use all caps
                        selected = selected.toUpperCase();
                    } else {
                        // otherwise just use so-far word (capitalized or mixed
                        // case)
                        selected = currentWord + selected.substring( currentWord.length() );
                    }
                }
            }
            // always perform caps request.
            if(isCapsLock()) {
                selected = selected.toUpperCase();
            } else if(isShifted()) {
                selected = capitalize(selected);
            }
            String curVal = getValue();
            // insert.
            int currentWordLength = currentWord.length();
            int cursorTextPosition = _cursorTextPosition;
            _cursorTextPosition += selected.length()-currentWordLength;
            setValueInternal(curVal.substring( 0, cursorTextPosition-currentWordLength )
                    +selected
                    +curVal.substring(cursorTextPosition));

            if(thenActive) {
                goActive();
            } else {
                getBApp().play( TiVo.SOUND_SPEEDUP1 );
                // then just do minimum updating.
                initActive();
            }
        }

        public void inactivate() {
            // does nothing
        }
    }
    
    private class KeyboardView extends BView {

        private BView[] _colViews;
        private BView[] _capsColViews;
        private int _colWidth;
        private int _charHeight;
        private BView _highlight;
        private int _charWidth;


        public KeyboardView(BView parent, int x, int y, int w, int h, boolean visible, Keyboard currentKeyboard) {
            super( parent, x, y, w, h, visible );
            makeHighlight();
			switch(currentKeyboard){
			case EMAIL:
				_selections = _email;
				break;
			case URL:
				_selections = _url;
				break;
			case PASSWORD:
				_selections = _default; //FIXME use a different layout?
				break;
			default:
				_selections = _default;
				break;
			}
			
			setSize(w, (int) (_selections.length*_charFontInfo.getHeight()));

            makeColumns();
        }

        private void makeHighlight() {
            _highlight = new BView( this, 0, 0, 10, 10, false );
            _highlight.setResource( Color.cyan );
            _highlight.setTransparency( .5f );
        }
        
		@Override
		public void remove(Resource animation) {
			_highlight.remove(animation);
			for(BView view : _colViews) {
				if(view.getResource() != null) view.getResource().remove();
				view.remove(animation);
			}
			for(BView view : _capsColViews) {
				if(view.getResource() != null) view.getResource().remove();
				view.remove(animation);
			}
			super.remove(animation);
		}

        private void makeColumns() {
            int maxCols = 0;
            for( int i = 0 ; i < _selections.length ; ++i ) {
                maxCols = Math.max( maxCols, _selections[i].length );
            }
            
            //TODO hmm..want to restrict view's size, but this is ending with nothing?
//            setSize( maxCols * _charWidth, _selections.length * _charHeight );
            
            _colViews = new BView[maxCols];
            _capsColViews = new BView[maxCols];
            _charHeight = (int)_charFontInfo.getHeight();
            _charWidth = (int)(_charHeight * //1.25
            		1.75);
            // calculate good width for the columns
			int maxWidth = _charHeight * 4;
			for (int col = 0; col < _colViews.length; ++col) {
				for (int i = 0; i < _selections.length; ++i) {
					if (col < _selections[i].length) {
						maxWidth = Math.max(maxWidth, _charFontInfo
								.measureTextWidth(_selections[i][col]));
					}
				}
			}
			int screenWidth = getWidth();
//					getScreen().getWidth()
//					- (2 * getBApp().getSafeTitleHorizontal());
			if (maxWidth * _colViews.length > screenWidth) {
				maxWidth = screenWidth / _colViews.length;
			}
            _colWidth = maxWidth;
            for( int col = 0 ; col < _colViews.length ; ++col ) {
                String colText = "";
                for( int i = 0 ; i < _selections.length ; ++i ) {
                    if( col < _selections[i].length ) {
                        colText += _selections[i][col];
                    }
                    colText += "\n";
                }
                int colX = 0;
                if(col > 0) {
					colX = _colViews[col - 1].getX()
							+ _colViews[col - 1].getWidth();
                }
                _colViews[col] = new BView( this, colX, 0, _colWidth,
                        _selections.length * _charHeight );
                _capsColViews[col] = new BView( this, colX, 0, _colWidth,
                        _selections.length * _charHeight, false );
                _colViews[col].setResource( getBApp().createText( _characterFontResource,
                        Color.white, colText ), Resource.RSRC_HALIGN_LEFT
                                                | Resource.RSRC_VALIGN_TOP );
                _capsColViews[col].setResource( getBApp().createText( _characterFontResource,
                        Color.white, colText.toUpperCase() ), Resource.RSRC_HALIGN_LEFT
                                                | Resource.RSRC_VALIGN_TOP );
            }
        }
        
        public void setCaps(boolean caps) {
            for(int i = 0 ; i < _colViews.length ; ++i) {
                _colViews[i].setVisible( !caps );
                _capsColViews[i].setVisible( caps );
            }
        }

        public boolean right() {
            if( _selections[_selectedRow].length <= _selectedColumn + 1 ) {
                return false;
            }
            ++_selectedColumn;
            updateHighlightCursor();
            return true;
        }

        public boolean left() {
            if( _selectedColumn <= 0 ) {
                return false;
            }
            --_selectedColumn;
            updateHighlightCursor();
            return true;
        }

        public boolean down() {
            if( _selectedRow + 1 >= _selections.length )
                return false;
            ++_selectedRow;
            if( _selections[_selectedRow].length <= _selectedColumn ) {
                _selectedColumn = _selections[_selectedRow].length - 1;
                if( _selectedColumn < 0 )
                    _selectedColumn = 0;
            }
            updateHighlightCursor();
            return true;
        }

        public boolean up() {
            if( _selectedRow <= 0 )
                return false;
            --_selectedRow;
            if( _selections[_selectedRow].length <= _selectedColumn ) {
                _selectedColumn = _selections[_selectedRow].length - 1;
                if( _selectedColumn < 0 )
                    _selectedColumn = 0;
            }
            updateHighlightCursor();
            return true;
        }
        
        private String _default[][] = new String[][]{
                new String[] {"s","sh","th","ch","es","ed","ly","ing","tion","ous"},
                new String[] {" ",",",".","'","\"","?","!","*","&",":"},
                new String[] {"a","e","i","o","u","y","/","@","(",")"},
                new String[] {"b","f","j","p","v","z","-","_","$","%"},
                //                                 |
                new String[] {"c","g","k","q","w","@","#","^","=","+"},
                new String[] {"d","h","l","r","x","*","\\","<",">",";"},
                //             |   |           |
                new String[] {"e","i","m","s","y","-","|","`","~"},
                new String[] {"f","j","n","t","z","_"}
        };
        private String _email[][] = new String[][]{
                new String[] {"@",".com",".org",".net",".us","ed","ly","ing","tion","ous"},
                new String[] {" ",",",".","'","\"","?","!","*","&",":"},
                new String[] {"a","e","i","o","u","y","/","@","(",")"},
                new String[] {"b","f","j","p","v","z","-","_","$","%"},
                //                                 |
                new String[] {"c","g","k","q","w","@","#","^","=","+"},
                new String[] {"d","h","l","r","x","*","\\","<",">",";"},
                //             |   |           |
                new String[] {"e","i","m","s","y","-","|","`","~"},
                new String[] {"f","j","n","t","z","_"}
        };
        private String _url[][] = new String[][]{
                new String[] {"http://","https://","www.",".com/",".org/",".net/",".us/"},//,"","","",""},
                new String[] {" ",",",".","'","\"","?","!","*","&",":"},
                new String[] {"a","e","i","o","u","y","/","@","(",")"},
                new String[] {"b","f","j","p","v","z","-","_","$","%"},
                //                                 |
                new String[] {"c","g","k","q","w","@","#","^","=","+"},
                new String[] {"d","h","l","r","x","*","\\","<",">",";"},
                //             |   |           |
                new String[] {"e","i","m","s","y","-","|","`","~"},
                new String[] {"f","j","n","t","z","_"}
        };
        private String _selections[][] = _default;
        private int _selectedRow;
        private int _selectedColumn;
        

        void selectFocused() {
            if( _selections.length > _selectedRow
                && _selections[_selectedRow].length > _selectedColumn ) {
                String selected = _selections[_selectedRow][_selectedColumn];
                type( selected );
                getBApp().play( TiVo.SOUND_SELECT );
            }
        }
        
        public void resetCursor() {
            _selectedRow = 0;
            _selectedColumn = 0;
            if(_selections[_selectedRow].length == 0) {
                _selectedRow = 1;
            }
            updateHighlightCursor();
        }
        
        private void updateHighlightCursor() {
//        	_selectedColumn * _charWidth
            _highlight.setLocation(_colViews[_selectedColumn].getX(), _selectedRow * _charHeight );
            int wid = _charWidth;
            if( _selectedRow >= 0 && _selections.length > _selectedRow
                && _selectedColumn >= 0 && _selections[_selectedRow].length > _selectedColumn ) {
                wid = measureTextWidth( _selections[_selectedRow][_selectedColumn] );
                if(_selections[_selectedRow][_selectedColumn].endsWith(" ")) {
                    wid += _fontTrailingSpaceAdditionalWidth;
                }
                wid = Math.max(wid, _charWidth);
            }
            _highlight.setSize( wid, _charHeight );
            _highlight.setVisible( true );
        }
        
        public void inactivate() {
            _highlight.setVisible( false );            
        }
        
    }
    
    String capitalize(String selected) {
        if( selected == null )
            return null;
        if( selected.length() > 1 ) {
            selected = selected.substring( 0, 1 ).toUpperCase()
                       + selected.substring( 1 ).toLowerCase();
        } else {
            selected = selected.toUpperCase();
        }
        return selected;
    }

//    If the HME application reports a minor version of at least 49, appropriate key code values are transmitted in keypress (and keyrepeat?) events.
//
//    Backspace is code 65
//    The enter key, when modified by the Alt or Windows key, is sent with code 70, otherwise it is sent as KEY_SELECT
//    Left Shift and Right Shift are codes 71 and 72
//    Left Alt and Right Alt (also Left and Right Windows key) 73 and 74
//    Capslock is code 75
//    ASCII characters 32 (space) through (tilde (~)) come in with a key code equal to the ASCII code plus 2^16 (65536). In other words key code 65568 (space) through 65662 (tilde). This does not include the numbers which send standard HME KEY codes for the numbers.
//    Other keys are assigned to standard HME key codes as listed below.
//    Keyrelease events, however, report with the below mappings using KEY_UNKNOWN as the key code.

    /**
	 * Backspace (picture of a left arrow on slide remote) button code.<BR>
	 * TODO codes 66-69 are unknown, currently
	 */
    public static final int KEY_BACKSPACE = 65;
    /**
	 * code returned by the RETURN key on the keyboard if it is modified by Alt
	 * (or Windows (or possibly "sym"(bol))). Normally that key sends
	 * KEY_SELECT.
	 */
    public static final int KEY_OPT_RETURN_MODIFIED = 70;
	/**
	 * Left Shift button code. You do not need to process this to handle
	 * capitalization, use KEY_ASCII_* to get the properly capitalized
	 * character.
	 */
    public static final int KEY_SHIFT_LEFT = 71;
    /**
	 * Right Shift button code. You do not need to process this to handle
	 * capitalization, use KEY_ASCII_* to get the properly capitalized
	 * character.
	 */
    public static final int KEY_SHIFT_RIGHT = 72;
    /**
	 * Left Alt (and Windows) key button code for keyboards - POSSIBLY the
	 * "sym"(bol) key for the Slide Remote (although there is only one, on the
	 * right side). You do not need to process this to handle symbols, use
	 * KEY_ASCII_* to get the properly capitalized character.
	 */
    public static final int KEY_OPT_ALT_LEFT = 73;
    /**
	 * Right Alt (and Windows) key button code for keyboards - POSSIBLY the
	 * "sym"(bol) key for the Slide Remote. You do not need to process this to
	 * handle symbols, use KEY_ASCII_* to get the properly capitalized
	 * character.
	 */
    public static final int KEY_OPT_ALT_RIGHT = 74;
	/**
	 * Capslock button code. You do not need to process this to handle
	 * capitalization, use KEY_ASCII_* to get the properly capitalized
	 * character.
	 */
    public static final int KEY_CAPSLOCK = 75;
	/**
	 * space character - beginning of range of values that are ascii characters
	 * plus KEY_ASCII_OFFSET. Note that numbers do not show up this way, only as
	 * KEY_NUM*
	 */
    public static final int KEY_ASCII_START = 65568;
	/**
	 * tilde character - end of range of values that are ASCII characters plus
	 * KEY_ASCII_OFFSET. Note that numbers do not show up this way, only as
	 * KEY_NUM*
	 */
    public static final int KEY_ASCII_END = 65662;
	/**
	 * For all characters between KEY_ASCII_START and KEY_ASCII_END (inclusive)
	 * subtract this value from the key code to get the ASCII character they
	 * represent
	 */
    public static final int KEY_ASCII_OFFSET = 65536; // 2^16
    
    public boolean handleSlideRemoteKeyPress(int code, long rawcode) {
//    	if(isSlideRemoteSupported()) {
		if (code >= KEY_ASCII_START
				&& code <= KEY_ASCII_END) {
    		// handle V49 KEY_ values 32-126 
			//TODO should numbers above 126 (~) be included?
			char slideRemoteCharacter = (char) (code - KEY_ASCII_OFFSET);
			if(slideRemoteCharacter <= 'z' && slideRemoteCharacter >= 'a') {
				// cancel capslock if we got out of synch
				setUnshifted();
			}
//    		System.out.println("SLIDE REMOTE! (not currently handled this way): \""+slideRemoteCharacter+"\"");
			String typed = String.valueOf(slideRemoteCharacter);
			if(type(typed)) {
				getBApp().play(TiVo.SOUND_UPDOWN);
			} else {
				getBApp().play(TiVo.SOUND_BONK);
			}
			// flag to not process key release events which are also getting sent :/
			_slideRemoteActive = true;
			return true;
    	}
		if(code == KEY_BACKSPACE) {
			if(backspace()) {
				getBApp().play(TiVo.SOUND_UPDOWN);
			} else {
				getBApp().play(TiVo.SOUND_BONK);
			}				
			return true;
		}
		// just give a visual cue about the capslock state as best we can guess it.
		if(code == KEY_CAPSLOCK) {
			if(isCapsLock()) {
				setUnshifted();
			} else {
				setCapsLock();
			}
		}
		//TODO could also process shift (left/right) but have to clear on key release and would cause lag, probably
//    	}
		return false;
	}

	boolean backspace() {
        if(_cursorTextPosition <= 0 ) return false;
        String val = getValue();
        int cursorTextPosition = _cursorTextPosition;
        --_cursorTextPosition;
        setValueInternal( val.substring(0, cursorTextPosition-1) +
                val.substring( cursorTextPosition ) );
        updateAll();
        return true;
    }

    boolean replay() {
        // go to previous word.
        if(_cursorTextPosition <= 0) {
            return false;
        } else {
            //TODO treat all non-word punctuation the same as space?
            // find space before current position
            int newPosition = _value.lastIndexOf( ' ', _cursorTextPosition-1 );
            // but treat concurrent spaces as a single space.
            while(newPosition > 0 && _value.charAt(newPosition-1) == ' ') {
                --newPosition;
            }
            // no spaces? go to beginning.
            if(newPosition < 0) {
                newPosition = 0;
            }
            _cursorTextPosition = newPosition;
        }
        updateAll();
        return true;
    }

    boolean advance() {
        //TODO go to next word?
        if(_cursorTextPosition < getValue().length()) {
            _cursorTextPosition = getValue().length();
        } else {
            _cursorTextPosition = 0;
        }
        updateAll();
        return true;
    }

    void updateAll() {
        updateCursor();
        updateCompletions();
    }
    
    public void setUnshifted() {
        _shifted = false;
        _capslock = false;
        updateShift();
    }

    public void setCapsLock() {
        _capslock = true;
        _shifted = true;
        updateShift();
    }

    public boolean isShifted() {
        return _shifted;
    }
    
    public boolean isCapsLock() {
        return _capslock;
    }

    public void setShifted() {
        _shifted = true;
        _capslock = false;
        updateShift();
    }
    
	private void shiftUp() {
		if (isShifted()) {
			if (isCapsLock()) {
				getBApp().play(TiVo.SOUND_SLOWDOWN1);
				setUnshifted();
			} else {
				getBApp().play(TiVo.SOUND_SPEEDUP3);
				setCapsLock();
			}
		} else {
			getBApp().play(TiVo.SOUND_SPEEDUP2);
			setShifted();
		}
	}

    /**
     * Set TyPo exit behaviors.
     * @param leftEdge Behavior when user goes beyond left edge: Blocked
     *        (default), Cancel, or Finished
     * @param rightEdge Behavior when user goes beyond right edge: Blocked,
     *        Cancel, or Finished (default)
     * @param finishedIncludesKeyboardEdge whether finished behavior includes
     *        hitting the edge of the keyboard (as well as the text cursor area)
     *        (default: true)
     * @param cancelIncludesKeyboardEdge whether cancel behavior includes
     *        hitting the edge of the keyboard (as well as the text cursor area)
     *        (default: false)
     * @param onFinishedForwardEdgeEvent whether to forward the cursor key event
     *        to enclosing views when it caused a "finished" behavior - i.e.
     *        finish-and-move-focus-right/left. (default: false)
     * @param onCancelForwardEdgeEvent whether to forward the cursor key event
     *        to enclosing views when it caused a "cancel" behavior - i.e.
     *        cancel-and-move-focus-right/left. (default: false)
     */
    public void setExitOptions(int leftEdge, int rightEdge, 
            boolean finishedIncludesKeyboardEdge, boolean cancelIncludesKeyboardEdge,
            boolean onFinishedForwardEdgeEvent, boolean onCancelForwardEdgeEvent)
    {
        _leftEdge = leftEdge;
        _rightEdge = rightEdge;
        _finishedIncludesKeyboardEdge = finishedIncludesKeyboardEdge;
        _cancelIncludesKeyboardEdge = cancelIncludesKeyboardEdge;
        _onFinishedForwardEdgeEvent = onFinishedForwardEdgeEvent;
        _onCancelForwardEdgeEvent = onCancelForwardEdgeEvent;
    }
    
    public int getLeftEdgeExit() {
        return _leftEdge;
    }
    
    public int getRightEdgeExit() {
        return _rightEdge;
    }
    
    private String _originalValue;
    
    public void cancelEntry() {
        setValue(_originalValue);
        goInactive();
        sendAction(_cancelAction);
    }

    public void finishEntry() {
        goInactive();
        sendAction(_finishedAction);
    }

    /**
     * Actions to post when the user is done with the TyPo keyboard entry.
     * @param finishedAction action to post if the user is finished entering
     *        text.
     * @param cancelAction action to post of the user cancels entering text.
     */
    public void setActions(Object finishedAction, Object cancelAction) {
        _finishedAction = finishedAction;
        _cancelAction = cancelAction;
    }
    
    private void sendAction(Object action) {
        if( action != null ) {
            postEvent( new BEvent.Action( TyPo.this, action ) );
        }
    }

    public Object getFinishedAction() {
        return _finishedAction;
    }
    
    public Object getCancelAction() {
        return _cancelAction;
    }

    
    BView _shiftView;
    Object _shiftTextFont = "default-16.font";
    private Resource _shiftText;
    private Resource _capsText;
    private Resource _noshiftText;
    private Resource _cursorShiftText;
    private String[] _highlightsToInactivate = new String[] {H_UP, H_DOWN, H_LEFT, H_RIGHT};

    /**
     *  indicate shift and capslock visually (upcase chars and extra for capslock)
     */
    private void updateShift() {
        _keyboardView.setCaps( _shifted );
        if(_capslock) {
            _shiftView.setResource(_capsText
                    , RSRC_TEXT_WRAP|RSRC_HALIGN_LEFT|RSRC_VALIGN_TOP                    );
        } else if(_shifted) {
            _shiftView.setResource(_shiftText
                    , RSRC_TEXT_WRAP|RSRC_HALIGN_LEFT|RSRC_VALIGN_TOP            );
        } else {
            _shiftView.setResource(_noshiftText
                    , RSRC_TEXT_WRAP|RSRC_HALIGN_LEFT|RSRC_VALIGN_TOP            );
        }
        // redraw completions with the current shift state.
        updateCompletions();
        
    }

    public boolean type(String text) {
        if(text == null) return false;
        if(isShifted()) {
            text = text.toUpperCase();
            if(!isCapsLock()) {
                setUnshifted();
            }
        }
        String val = getValue();
		if (this.maxLength > 0 && val.length() >= this.maxLength) {
			return false;
		}
		int cursorTextPosition = _cursorTextPosition;
        _cursorTextPosition+=text.length();
        setValueInternal( val.substring(0,cursorTextPosition) + text +
                val.substring( cursorTextPosition ) );
        updateAll();
        return true;
    }
    
    void initActive() {
        if(_completionView == null) {
            // create the two views
            _completionView = createCompletionView();
            _keyboardView = createKeyboardView();
            _shiftView = new BView((BView)getParent(), getX()+getWidth(), getY(),100,getHeight()*8);
            String mainHelp = "<<: backspace\n>>: space\nreplay: prev word";
            String keyboardHelp = "1-9: completion\nenter/pause: back\n"+mainHelp;
            _cursorShiftText = createText(_shiftTextFont, Color.yellow, "select: end\nup: completions\ndown: keyboard\n"+mainHelp+"\n(thumbs: change letter's case)");
            _shiftText = createText( _shiftTextFont, Color.yellow, "Shift (thumbs: change)\n"+keyboardHelp);
            _capsText = createText( _shiftTextFont, Color.yellow, "LOCK (thumbs: change)\n"+keyboardHelp);
            _noshiftText = createText( _shiftTextFont, Color.yellow, keyboardHelp+"\n(thumbs: Shift/LOCK)");
            updateShift();
        }
        
        if(!_keyboardView.getVisible()) {
            _cursorTextPosition = getValue().length();

            Resource showAnim = getResource("*500");
            // if needed, move everything to be visible, storing original locations.
            int screenActionMaxX = getApp().getRoot().getWidth()-SAFE_TITLE_H;
            int screenActionMaxY = getApp().getRoot().getHeight()-SAFE_TITLE_V;
            int completionHeight = _completionView.getHeight();
            if(completionHeight < _completionList.getRowHeight()*10) {
                completionHeight = _completionList.getRowHeight()*10;
                _completionView.setSize( _completionView.getWidth(), completionHeight );
                _completionList.setSize( _completionView.getWidth(), completionHeight );
            }
            int completionWidth = _completionView.getWidth();
            if(_originalWidth < 0 && 
                    (completionHeight > this.getY()+SAFE_TITLE_V ||
                            _completionView.getY() < SAFE_TITLE_V ||
                            completionWidth+_completionView.getX()> screenActionMaxX ||
                            _keyboardView.getHeight()+_keyboardView.getY() > screenActionMaxY
                    )) {
                _originalX = this.getX();
                _originalY = this.getY();
                //TODO adjust width if necessary
                _originalWidth = this.getWidth();
                
                int xDiff = SAFE_TITLE_H-getX();
                int yDiff = completionHeight+SAFE_TITLE_V-getY();
                _completionView.setLocation(_completionView.getX()+xDiff,_completionView.getY()+yDiff,showAnim);
                this.setLocation(SAFE_TITLE_H,completionHeight+SAFE_TITLE_V,showAnim);
                _shiftView.setLocation( _shiftView.getX()+xDiff, _shiftView.getY()+yDiff,showAnim);
                _keyboardView.setLocation(_keyboardView.getX()+xDiff,_keyboardView.getY()+yDiff,showAnim);
                // move the main button highlight, too.
                this.getHighlights().refresh(showAnim);
//            } else {
//                _originalX = -1;
//                _originalY = -1;
//                _originalWidth = -1;
            }
            if(isUsingCompletions && currentKeyboard != Keyboard.PASSWORD) {
            	_completionView.setVisible( true );
            } else {
            	_completionView.setVisible( false );
            }
            _keyboardView.setVisible( true );
            _shiftView.setVisible( true );
            // activate the two views
        }
        
        updateAll();        
        
//        _completionView.clearResource();

        activate();
    }

    private void removeResource(View view) {
        if(view.getResource() != null) 
            view.getResource().remove();
        view.clearResource();
    }
    private KeyboardView createKeyboardView() {
        KeyboardView bView = new KeyboardView((BView)getParent(), 
                getX(), getY()+getHeight(), 
                _toolWidth, 400, false, currentKeyboard);
        bView.setResource( new Color(0, 0, 200, 128) );
        
        return bView;
    }
    private CompletionView createCompletionView() {
        
        CompletionView bView = new CompletionView((BView)getParent(), 
                getX(), getY()-(getHeight()*10), 
                _toolWidth, getHeight()*10, false);
//        bView.setResource( new Color(0, 0, 200, 128) );
        
        _completionList = new GenericList(bView, 
              getX(), 0, 
              _toolWidth, getHeight()*10,
              getHeight());
        _completionList.setFont( _characterFontString );
        _completionList.setBarAndArrows( BAR_DEFAULT, BAR_DEFAULT, 
                COMPLETION_ABORTED, COMPLETION_SELECT_RIGHT );
        _completionList.setSelectAction( COMPLETION_SELECT );
        
        _completionList.setResource(new Color(0, 0, 200, 128));
        
        return bView;
    }
    /**
     * adjust main TyPo view to indicate it is active.
     */
    private void activate() {
        updateCursor();
        setResource(getResource("0x4444FF"));
        _cursor.setVisible( true );
    }
    /**
     * adjust main TyPo view to indicate it is not active
     */
    private void inactivate() {
        this.clearResource();
        _cursorTextPosition = 0;
        updateCursor();
        _cursor.setVisible( false );
        _textHighlighter.setVisible( false );
    }
    
    /**
     * Handle the codes when a keyboard is in use as of 11.0h (for S3/HD) or
     * 14.5 (for Premiere).  If sw version 0.49 or above, should not be used (should be using slideRemote handling instead), so attempts to disable itself.
     * @param rawcode
     * @return true if the code was understood / handled.
     */
    boolean handleKeyboard(long rawcode) {
    	if(isSlideRemoteMode()) {
    		return true;
    	}
        String typed = getKeyboardString(rawcode);
        if(typed != null) {
        	//rawcode-based strings are the low-tech version without case info
        	if(isShifted()) {
        		typed = typed.toUpperCase();
        	} else {
        		typed = typed.toLowerCase();
        	}
			if(type(typed)) {
				getBApp().play(TiVo.SOUND_UPDOWN);
			} else {
				getBApp().play(TiVo.SOUND_BONK);
			}
            return true;
//        } else {
//            int keycode = Utility.getKeyboardCode(rawcode);
//            if(keycode > 0) {
//                return true;
//            }
        }
        return false;
    }

    private boolean _slideRemoteActive = false;
    private boolean isSlideRemoteMode() {
    	return _slideRemoteActive;
	}

	private static char[] keyboardRawcodeTable = new char[] {
        //0-25
        'A','B','C','D','E','F','G','H','I','J','K','L','M',
        'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
        '-',//26
        '=',
        '[',
        ']',
        '\\',
        ';',
        '\'',
        ',',
        '.',
        '/',
        '`',
        ' ',//37 (space)
        '\t'//38
    };
    /**
     * get KEY_UNKNOWN REPEAT/RELEASE event translations (most aren't sent on PRESS)
     * @param rawcode
     * @return the string that corresponds to the rawcode, or null.
     */
    public static String getKeyboardString(long rawcode) {
        int adjustedCode = (int)(((rawcode & 0xff00) >> 8) - 0x3c);
        if( adjustedCode >= 0 && adjustedCode < keyboardRawcodeTable.length ) {
            return String.valueOf( keyboardRawcodeTable[adjustedCode] );
        } else {
            return null;
        }
    }
    
    boolean _goActiveOnFontInfo = false;
    void goActive() {
        if(_charFontInfo == null) {_goActiveOnFontInfo = true; return;}
        
        // save value for cancelling if coming from inactive
        if(getCurrentStrategy() == _inactiveStrategy) {
            _originalValue = getValue();
        }
        
        // set highlights to not have any arrows.
        for(int i = 0 ; i < _highlightsToInactivate.length ; ++i) {
            String name = _highlightsToInactivate[i];
            if(getHighlights().get( name ) != null) {
//                    // TODO ARGH! getVisible doesn't exist!
//                    _inactiveHighlights.set( getHighlights().get( name ).getVisible() );
                getHighlights().get( name ).setVisible( H_VIS_FALSE );
            }
        }
//        getHighlights().refresh(null);
        
        setKeyStrategy(_activeStrategy);

        initActive();
        if(_keyboardView != null) {
            _keyboardView.inactivate();
        }
        if(_completionView != null) {
            _completionView.inactivate();
        }
        setUnshifted();
        if(_shiftView != null) {
            _shiftView.setResource( _cursorShiftText, RSRC_TEXT_WRAP|RSRC_HALIGN_LEFT|RSRC_VALIGN_TOP );
        }
        updateTextHighlight();
        _textHighlighter.setVisible( true );
        this.setFocus();
        getBApp().play(TiVo.SOUND_SELECT);
        
    }
    
    void goInactive() {
        // restore highlights
        for(int i = 0 ; i < _highlightsToInactivate.length ; ++i) {
            String name = _highlightsToInactivate[i];
            if(getHighlights().get( name ) != null) {
//               setVisible( _inactiveHighlights[i] );
                getHighlights().get( name ).setVisible( H_VIS_FOCUS );
            }
        }
//        getHighlights().refresh(null);
        
        setKeyStrategy(_inactiveStrategy);
        
        Resource hideAnim = getResource("*250");
        if(_originalWidth > -1) {
            _completionView.setLocation(_originalX,_originalY-_completionView.getHeight(),hideAnim);
            this.setLocation(_originalX,_originalY,hideAnim);
            _shiftView.setLocation( _originalX+getWidth(), _originalY,hideAnim);
            _keyboardView.setLocation(_originalX,_originalY+this.getHeight(),hideAnim);
            // move the main button highlight, too.
            this.getHighlights().refresh(hideAnim);

            _originalX = -1;
            _originalY = -1;
            _originalWidth = -1;
        }
        if(_keyboardView.getVisible()) {
            // inactivate the two views
            _completionView.setVisible( false );
            _keyboardView.setVisible( false );
            _shiftView.setVisible( false );
        }
        
        getBApp().play( TiVo.SOUND_DESELECT );
        
        if(currentKeyboard != Keyboard.PASSWORD) {
        	addNewWords(getValue());
        }
        
        inactivate();

    }
    /**
     * add new words to user dictionary
     * @param value
     */
    private void addNewWords(String value) {
        String words[] = value.split( "[^a-zA-Z'-]+" );
        for(String word : words) {
            // if this word is not known...
            if(WordList.getInstance().getSuggestions( word ) == null || 
                    ! WordList.getInstance().getSuggestions( word ).contains( word )) {
                // even if we ignore its case... (don't add capitalized known words)
                if(WordList.getInstance().getSuggestions( word.toLowerCase() ) == null || 
                        ! WordList.getInstance().getSuggestions( word.toLowerCase() ).contains( word.toLowerCase() )) {
                    //TODO arbitrary minimum word length.
                    //TODO maybe words should include adjacent, non-trailing dots?
                    if(word.length() > 2) {
                        WordList.getInstance().addWord( word );
                        // add it to use dictionary
                        addUserWord(word);
                    }
                }
            }
        }
    }

	private static boolean isUserWordListAllowed = true;
	public static void setUserWordListAllowed(boolean allow) {
		isUserWordListAllowed = allow;
	}
	
    static File _userWordFile;
    private void addUserWord(String word) {
        if(isUserWordListAllowed) {
	        try {
	            BufferedWriter writer = new BufferedWriter(new FileWriter(_userWordFile, true));
	            writer.write( word );
	            writer.newLine();
	            writer.close();
	        } catch(IOException e) {
	            System.out.println("writing user word: "+e);
	            // unable to write to user word file
	            // FIXME log this somehow.
	        }
        }
    }

    void goCompletion() {
    	if(isUsingCompletions && currentKeyboard != Keyboard.PASSWORD) {
	        setKeyStrategy(_completionStrategy);
	
	        _textHighlighter.setVisible( true );
	        updateShift();
	        GenericList completionList = _completionList;
	        completionList.setFocus( completionList.size()-1, false );
	//        _completionView.setResource( getResource("0x4444FF") );
	        getBApp().play(TiVo.SOUND_DESELECT);
    	}
    }
    void goKeyboard() {
        setKeyStrategy(_keyboardStrategy);

        _textHighlighter.setVisible( true );
        _keyboardView.resetCursor();
        updateShift();
        getBApp().play(TiVo.SOUND_DESELECT);
    }

    private BView _textFrame;
    private BView _textHighlighter;
    private BView _text;
    String _value;
    private BView _cursor;
    private KeyStrategy _currentStrategy;
    private KeyStrategy _inactiveStrategy;
    private KeyStrategy _activeStrategy;
    private KeyStrategy _completionStrategy;
    private KeyStrategy _keyboardStrategy;
    private String _characterFontString;
    Resource _characterFontResource;
    int _cursorTextPosition;
    FontInfo _charFontInfo;
    int _fontTrailingSpaceAdditionalWidth;
    private int _toolWidth;
    private Color _cursorColor = Color.green.brighter();
    private Color _cursorInactiveColor = Color.green.darker();

    //TODO understand this better...tried putting files at typo level, but it couldn't find them.
    private static String[] frequentListFiles = new String[] {
        "/FREQ.TXT"};
    private static String[] wordListFiles = new String[] {
        "/COMMON.TXT",
        "/CROSSWD.TXT","/CRSWD-D.TXT",
        "/SINGLE.TXT","/COMPOUND.TXT","/ACRONYMS.TXT",
        "/NAMES.TXT","/NAMES-M.TXT","/NAMES-F.TXT","/PLACES.TXT",
        "/OFTENMIS.TXT"
        };
    
	private static boolean isInited;    
    private static void initWordList() {
    	if(!isInited) {
	        // allow phrases (will include "Decreasing Order of Occurrence" from FREQ.TXT)
	        WordList.getInstance().setSkipLinesWithSpaces( false );
	        for( String file : frequentListFiles ) {
	            InputStream fileStream = TyPo.class.getResourceAsStream( file );
	            if( fileStream != null ) {
	                try {
	                    WordList.getInstance().load( fileStream );
	                } catch(Exception e) {
	                    // Allowable
	                    System.err.println("loading "+file+": "+e);
	                    e.printStackTrace();
	                }
	            }// else System.err.println(file+" not found");
	        }
	        if(isUserWordListAllowed) {
		        File userWordHome = new File(System.getProperty( "user.home", "/" ));
		        _userWordFile = new File(userWordHome,"TyPo_for_TiVo_User_Words.txt");
		        try {
		            WordList.getInstance().load(_userWordFile);
		        } catch(IOException e) {
		            // OK... may not have one.
		            System.err.println("loading user file: "+e);
		            e.printStackTrace();
		        } catch(Exception e) {
		            // Allowable
		            System.err.println("loading user file: "+e);
		            e.printStackTrace();
		        }
	        }
	        for( String file : wordListFiles ) {
	            InputStream fileStream = TyPo.class.getResourceAsStream( file );
	            if( fileStream != null ) {
	                try {
	                    WordList.getInstance().load( fileStream );
	                } catch(Exception e) {
	                    // Allowable
	                    System.err.println("loading "+file+": "+e);
	                    e.printStackTrace();
	                }
	            }// else System.err.println(file+" not found");
	        }
	        if(WordList.getInstance().getWordCount() == 0) {
	            // This is a real error... of sorts.
	            // not really...can still build a personal dictionary and
	            // can still work as a basic keyboard.
	        }
	        isInited=true;
    	}
    }
    public TyPo(BView parent, int x, int y, int width, int height, boolean visible) {
    	this(parent, x, y, width, height, visible, -1);
    }
    
    /**
     * 
     * @param parent
     * @param x
     * @param y
     * @param width
     * @param height
     * @param visible
     * @param protocolVersion the "version" parameter passed in during {@link com.tivo.hme.sdk.Application#setContext(IContext, int)}
     */
    public TyPo(BView parent, int x, int y, int width, int height, boolean visible, int protocolVersion) {
        super( parent, x, y, width, height, visible );
		int VERSION_0_49 = (0 << 8) | 49; // no sdk - Slide Remote key events        
        if(protocolVersion >= VERSION_0_49) {
			_slideRemoteActive = true;
		}
        _value = "";
        initWordList();
        
        // at least 300 for the keyboard/completion views.
        _toolWidth = Math.max(300,width);
        
        // this step gives us the font resource, allowing us to capture the
        // fontInfo event so we can calculate sizes.
        
        // we have to create the font resource by hand because createText
        // doesn't set the flags to make sure we get our font info.
        _characterFontResource = createFont( 
                TiVo.FONT_DEFAULT, FONT_PLAIN, 
                // size is the available pixels for the text
                //FIXME should be adjustable based on height probably.
                height,
                FONT_METRICS_BASIC | FONT_METRICS_GLYPH );
        _characterFontString = "default-"+height+".font";
        _characterFontResource.addHandler( this );
        flush();
        
        // When the font info event comes in, we'll finish initializing.
        // with initTyPo
    }
    
    //FIXME doesn't seem to work? on testapp with nothing else to focus, anyhow.
    private boolean activateOnFocus = false;
	private boolean isUsingCompletions = true;
	private Keyboard currentKeyboard;
	private List<String> defaultSuggestions;
	private int maxLength;
    /**
     * Set whether or not the widget automatically activates when it receives
     * focus, otherwise the user must press SELECT or ENTER to activate it.
     * @param activateOnFocus
     */
    public void setActivateOnFocus(boolean activateOnFocus) {
        this.activateOnFocus = activateOnFocus;
    }
    public boolean isActivateOnFocus() {
        return this.activateOnFocus;
    }
    @Override
    public boolean handleFocus(boolean focus) {
//        System.out.println("focus "+focus+" is inactive "+(getCurrentStrategy() == _inactiveStrategy));
        if(focus && isActivateOnFocus() && getCurrentStrategy() == _inactiveStrategy) {
            goActive();
            return true;
        }
        return super.handleFocus( focus );
    }
    
    public void activateNow() {
    	goActive();
    }
    
    /**
     * Catch the FontInfo event we registered for with our font resource.
     *
     * {@inheritDoc}
     * @see com.tivo.hme.bananas.BView#handleEvent(com.tivo.hme.sdk.HmeEvent)
     */
    public boolean handleEvent(HmeEvent event) {
        // save the font info for calculations.
        if(event.getOpCode() == EVT_FONT_INFO) {
            _charFontInfo = ((FontInfo)event);
            // now that we know font sizes, we can finish initializing.
            initTyPo();
            if(_goActiveOnFontInfo) {
                goActive();
                _goActiveOnFontInfo = false;
            }
        }
        return super.handleEvent( event );
    }    

    public static final int KEY_OPT_A = 56;
    public static final int KEY_OPT_B = 57;
    public static final int KEY_OPT_C = 58;
    public static final int KEY_OPT_D = 59;
    
    public static final int LIMIT_TEXT_RENDER_WIDTH=1920;
    
    private void initTyPo() {
        _textFrame = new BView( this, 0, 0, getWidth(), getHeight() );
        // for bgcolor-based highlighting of a region of the text
        _textHighlighter = new BView(_textFrame, 0, 0, 0, _textFrame.getHeight(), false);
        _textHighlighter.setResource( new Color(0,0,0,128));
        // hold the actual text - larger than parent to allow smooth scrolling.
        _text = new BView(_textFrame, 0, 0, LIMIT_TEXT_RENDER_WIDTH, getHeight());
        _cursor = new BView(_textFrame, 0, 1, 3, getHeight()-2, false);
        //TODO better cursor resource.
        _cursor.setResource( _cursorColor );

        _inactiveStrategy = new InactiveStrategy();
        _activeStrategy = new ActiveStrategy();
        _completionStrategy = new CompletionStrategy();
        _keyboardStrategy = new KeyboardStrategy();
        setKeyStrategy(_inactiveStrategy);
        
        // make sure any interim setValues have their associated updates completed.
        setValue(_value);
    }
    
    /**
     * Get the current value contained by this widget.
     * @return the current String represented by this widget (never null).
     */
    public String getValue() {
        return _value;
    }

    /**
     * Initialize or change the widget's value with this method. 
     * @param value (if null, replaced with an empty String)
     */
    public void setValue(String value) {
        setValueInternal( value );
        if( _cursorTextPosition > 0 ) {
            _cursorTextPosition = Math.min( _cursorTextPosition, _value.length() );
        } else {
            _cursorTextPosition = _value.length();
        }
        updateAll();
    }
    
    void setValueInternal(String value) {
        if(value == null) value = "";
        _value = value;
        updateText();
    }
    
    private void updateText() {
        if( _text != null ) {
            removeResource( _text );
            String displayValue = getDisplayValue();
            _text.setResource( createText( _characterFontResource, "0xFFFFFF", displayValue ), 
                    Resource.RSRC_HALIGN_LEFT | Resource.RSRC_VALIGN_CENTER);
        }
    }

	private String getDisplayValue() {
		String displayValue = _value;
		if (currentKeyboard == Keyboard.PASSWORD && _value.length() > 1) {
			//TODO needs to do this update on cursor moves
			displayValue = "";
			// show ***a|
			for (int i = 0; i < _cursorTextPosition-1; ++i) {
				displayValue += "*";
			}
			if (_cursorTextPosition-1 < _value.length()) {
				if(_cursorTextPosition > 0) {
					displayValue += _value.charAt(_cursorTextPosition-1);
				}
		    	// show ***a|***
				for(int i = _cursorTextPosition ; i < _value.length() ; ++i) {
					displayValue += "*";
				}
			}
		}
		return displayValue;
	}

    /**
     * Basically redraw both views
     */
    void updateCompletions() {
    	// also don't use completions if you're typing a password
    	if(isUsingCompletions && currentKeyboard != Keyboard.PASSWORD) {
    		updateTextHighlight();
    		updateCompletionsList();
    		updateLetters();
    	}
    }
    /**
     * Custom measureTextWidth to deal with an issue with the standard one.
     * @param text
     * @return
     */
    private int measureTextWidth(String text) {
        if(text == null || text.length() == 0)
            return 0;
        // measureTextWidth short-changes some trailing characters based on the
        // assumption their bounding width value is greater than or equal to the
        // advance, but apparently that's not necessarily true.
        int highlightStart = _charFontInfo.measureTextWidth( text );
        GlyphInfo lastGlyph = _charFontInfo.getGlyphInfo( text.charAt( text.length()-1 ) );
        if(lastGlyph.getAdvance() > lastGlyph.getBoundingWidth()) {
            highlightStart += (lastGlyph.getAdvance()-lastGlyph.getBoundingWidth());
        }
        return highlightStart;
    }
    private void updateTextHighlight() {
        // highlight current word since it will be replaced by completion.
        if(_charFontInfo != null && _textHighlighter != null) {
        	if(currentKeyboard == Keyboard.PASSWORD) {
        		_textHighlighter.setVisible(false);
        		_textHighlighter.setSize(0,_textHighlighter.getHeight());
        	} else {
        		int highlightStart = measureTextWidth( getDisplayValue().substring(0,_cursorTextPosition) );
        		int highlightWidth = measureTextWidth( getCurrentWord() );
        		highlightStart -= highlightWidth;
        		// adjust by _text's current scroll position.
        		_textHighlighter.setBounds( highlightStart + _text.getX(), _textHighlighter.getY(), highlightWidth, _textHighlighter.getHeight(),
        				getResource("*100") );
        		_textHighlighter.setVisible(true);
        	}
        }
    }

    private void updateLetters() {
//        _keyboardView.setFirstLetters(letters);
        // TODO idea was to include likely next letters as first row of keyboard, but seems unnecessary
        
    }

    private void updateCompletionsList() {
        BList completionList = _completionList;
        if(completionList != null) {
            //try to re-focus on focusedString after rebuilding.
            // save the previously focused string in hopes of maintaining that focus.
            int focused = completionList.getFocus();
            String focusedString = "";
            if( focused > -1 ) {
                focusedString = (String)completionList.get( focused );
                focusedString = focusedString.substring( focusedString.lastIndexOf( " " ) + 1 );
            }
            
            completionList.clear();
            // completion list positioned based on begin of word, not end.
            int completionX = _cursor.getX()+_textFrame.getX()-
            25-//FIXME calculate offset for list to text
            measureTextWidth( getCurrentWord() );
            // don't go beyond edge of screen, though.
            if(completionX+_completionList.getWidth() > getScreen().getWidth()) {
                completionX = getScreen().getWidth()-_completionList.getWidth();
            }
            _completionView.setLocation(
                    completionX,
                    _completionView.getY(), getResource("*250") );
            
            String word = getCurrentWord();

			List<String> suggestions = null;
			if (word.length() == 0) {
				suggestions = getDefaultSuggestions();
			}
			if (suggestions == null || suggestions.size() == 0) {
				suggestions = WordList.getInstance().getSuggestions(word);
			}
            // try an uncapitalized version for suggestions.
            if(suggestions == null && word.length() > 0 && !word.toLowerCase().equals( word )) {
                suggestions = WordList.getInstance().getSuggestions(word.toLowerCase());
            }
            if(suggestions != null) {
                suggestions.remove(word);
                suggestions.remove(word.toLowerCase());
                if(suggestions.size() == 0) {
                    completionList.getScreen().setFocus(TyPo.this);
                }
                Vector<String> vector = new Vector<String>();
                // prefix suggestions for readability and shortcuts.
                int i = 1;
                for(String sug : suggestions) {
                    String sug2;
                    // apply current caps state to list items.
                    String sugCaps = sug;
                    if(isCapsLock()) {
                        sugCaps = sug.toUpperCase();
                    } else if(isShifted()) {
                        sugCaps = capitalize(sug);
                    }
                    if(i < 10) {
                        sug2 = "   "+i+": "+sugCaps;
                    } else {
                        sug2 = "   "+sugCaps;
                    }
                    vector.add(sug2);
                    // save the new definition of the focused string.
                    if(sug.equals(focusedString)) {
                        focusedString = sug2;
                    }
                    ++i;
                }
                // put Most Frequently Used at bottom - the first thing to select.
                Collections.reverse( vector );
                completionList.add( vector );
                // if the list is longer than view, make sure the last item is
                // in view
                completionList.setTop( Math.max( 0, completionList.size()
                                                    - completionList.getNVisibleRows() ) );
                // keep the list bottom-justified.
                if(completionList.size() < completionList.getNVisibleRows()) {
                    _completionView.setTranslation( 0,
                            (completionList.getNVisibleRows() - completionList.size())
                                    * completionList.getRowHeight() );
                } else {
                    _completionView.setTranslation( 0, 0 );
                }

                if(_currentStrategy == _completionStrategy) {
                    // if we can find the previously focused string again, focus it.
                    int focusMe = vector.indexOf( focusedString );
                    if(focusMe != -1) {
                        completionList.setFocus( focusMe, false );
                    }
                    // sometimes stuck in completion mode when nothing is focused,
                    // so focus lowest item at least
                    else {
                        completionList.setFocus(completionList.size()-1, false);
                    }
                }
                
            } else {
                completionList.getScreen().setFocus(TyPo.this);
            }
            if(completionList.size() == 0) {
                // force BList to re-evaluate whether it should have any
                // highlight bar displayed.
                completionList.add( "" );
                completionList.clear();
            } else {
                completionList.refresh();
            }
        }
    }

    public List<String> getDefaultSuggestions() {
		return defaultSuggestions;
	}

	/**
	 * Set the list of suggestions to use when there is no current partial word
	 * 
	 * @param defaults
	 */
    public void setDefaultSuggestions(List<String> defaults) {
    	defaultSuggestions = defaults;
    }
    
    public int getMaxLength() {
    	return this.maxLength;
    }
    public void setMaxLength(int max) {
    	this.maxLength = max;
    }

	String getCurrentWord() {
        int lastSpace = 0;
        if( _cursorTextPosition > 0 ) {
            lastSpace = _value.lastIndexOf( ' ', _cursorTextPosition - 1 );
            if( lastSpace < _cursorTextPosition ) {
                ++lastSpace; // -1=0, others = char after space
            }
        }
        String word = _value.substring(lastSpace, _cursorTextPosition);
        while( word.length() > 0 && !word.matches( "^[a-zA-Z-']+" ) ) {
            word = word.substring( 1 );
        }
        return word;
    }

    /**
     * make sure cursor position is visible in text, make sure cursor is located
     * at that position relative to the visible text
     */
    private void updateCursor() {
        if( _cursor != null ) {
            String positionString = getDisplayValue().substring( 0, _cursorTextPosition );
            int x = measureTextWidth( positionString );
            // scroll the text so cursor is visible.
            Resource smoothScroll = getResource("*200");
            if(x > getWidth()) {
                _text.setLocation(getWidth()-x-20, _text.getY(),smoothScroll);
                _cursor.setLocation( x+(getWidth()-x-20), _cursor.getY() ,smoothScroll);
            } else {//if(x < getWidth()) {
                _text.setLocation(0, _text.getY(),smoothScroll);
                _cursor.setLocation( x, _cursor.getY() ,smoothScroll);
            }
        }
    }
    

    @Override
    public boolean handleKeyPress(int code, long rawcode) {
        if(_currentStrategy != null && _currentStrategy.handleKeyPress(code, rawcode))
            return true;
        if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy) {
            return super.handleKeyPress( code, rawcode );
        } else {
            // Eat everything else if we're active.
            return false;
        }
    }
    @Override
    public boolean handleKeyRelease(int code, long rawcode) {
        if(_currentStrategy != null && _currentStrategy.handleKeyRelease(code, rawcode))
            return true;
        if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy) {
            return super.handleKeyRelease( code, rawcode );
        } else {
            // Eat everything else if we're active.
            return false;
        }
    }
    @Override
    public boolean handleKeyRepeat(int code, long rawcode) {
        if(_currentStrategy != null && _currentStrategy.handleKeyRepeat(code, rawcode))
            return true;
        if(getCurrentStrategy() == null || getCurrentStrategy() == _inactiveStrategy) {
            return super.handleKeyRepeat( code, rawcode );
        } else {
            // Eat everything else if we're active.
            return false;
        }
    }

    private void setKeyStrategy(KeyStrategy strategy) {
        _currentStrategy = strategy;
        if( _cursor != null ) {
            if( getCurrentStrategy() == _activeStrategy ) {
                _cursor.setResource( _cursorColor );
            } else {
                _cursor.setResource( _cursorInactiveColor );
            }
        }
        
    }
    
    KeyStrategy getCurrentStrategy() {
        return _currentStrategy;
    }


    
    public static class TestApp extends BApplication {
        private TyPo _typo;

        public TestApp() {
            super();
        }
        @Override
        public void init(IContext context)
            throws Exception
        {
            super.init( context );
            BScreen screen = new BScreen( this );
            _typo = new TyPo(screen.getNormal(), SAFE_TITLE_H, SAFE_TITLE_V, 100, 20, true);
            // test - should immediately activate
//            _typo.setActivateOnFocus( true );
            _typo.setBarAndArrows( -10, _typo.getWidth()+10, "left", "right", "up", "down", false );
            _typo.setValue( "the" );
            _typo.setActions( "finished", "cancelled" );
            _typo.setExitOptions( EXIT_FINISHED, EXIT_FINISHED, true, false, true, false );
            screen.setFocusDefault( _typo );
            push( screen, TRANSITION_FADE );
        }
        
        @Override
        public boolean handleAction(BView view, Object action) {
            System.out.println("action: "+action);
            return super.handleAction( view, action );
        }
    }

	public void setKeyboard(Keyboard keyboard) {
		if(keyboard != currentKeyboard) {
			currentKeyboard = keyboard;
			if(_keyboardView != null) {
				destroyKeyboardView(_keyboardView);
				_keyboardView = createKeyboardView();
			}
			if(getCurrentStrategy() != _inactiveStrategy) {
				goInactive();
				goActive();
			}
		}
	}
	
	private void destroyKeyboardView(KeyboardView keyboardView) {
		keyboardView.remove();
	}

	public void setUseCompletions(boolean use) {
		this.isUsingCompletions = use;
	}
    
}
